"""
===============
Exporting JPEGs
===============

This example shows how to export image data as JPEG file, using the 'jpg'
protocol.

The 'jpg' protocol accepts additional protocol arguments, like color
table, color scaling or pixel binning. For a list of available color
tables, see http://jsoc.stanford.edu/ajax/exportdata.html and select the
JPEG protocol.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

# Arguments for 'jpg' protocol
jpg_args = {
    "ct": "aia_304.lut",  # color table
    "min": 4,  # min value
    "max": 800,  # max value
    "scaling": "log",  # color scaling
    "size": 2,  # binning (1 -> 4k, 2 -> 2k, 4 -> 1k)
}

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: "Series[timespan][wavelength]{data segments}"

qstr = "aia.lev1_euv_12s[2012-08-31T19:48:01Z][304]{image}"
print(f"Data export query:\n  {qstr}\n")

# Submit export request using the 'jpg' protocol with custom protocol_args
print("Submitting export request...")
result = client.export(qstr, protocol="jpg", protocol_args=jpg_args, email=email)

# Print request URL.
print(f"\nRequest URL: {result.request_url}")
print(f"{int(len(result.urls))} file(s) available for download.\n")

# Download selected files.
result.download(out_dir)
print("Download finished.")
print(f"\nDownload directory:\n  {out_dir.resolve()}\n")



import pytest

import drms
from drms.config import ServerConfig


def test_client_init_defaults():
    c = drms.Client()
    assert isinstance(c._server, ServerConfig)
    assert c._server.name.lower() == "jsoc"
    assert c.email is None


@pytest.mark.parametrize("server_name", ["jsoc", "kis"])
def test_client_registered_servers(server_name):
    c = drms.Client(server_name)
    assert isinstance(c._server, ServerConfig)
    assert c._server.name.lower() == server_name
    assert c.email is None


def test_client_custom_config():
    cfg = ServerConfig(name="TEST")
    c = drms.Client(cfg)
    assert isinstance(c._server, ServerConfig)
    assert c._server.name == "TEST"


def test_repr():
    assert repr(drms.Client()) == "<Client: JSOC>"
    assert repr(drms.Client("kis")) == "<Client: KIS>"



import os
import re
import time
import shutil
import socket
from pathlib import Path
from collections import OrderedDict
from urllib.error import URLError, HTTPError
from urllib.parse import urljoin
from urllib.request import urlopen

import numpy as np
import pandas as pd

from drms import logger
from drms.utils import create_request_with_header
from .exceptions import DrmsExportError, DrmsOperationNotSupported, DrmsQueryError
from .json import HttpJsonClient
from .utils import _extract_series_name, _pd_to_numeric_coerce, _split_arg

__all__ = ["Client", "ExportRequest", "SeriesInfo"]


class SeriesInfo:
    """
    DRMS series details.

    Use :func:`Client.info` to create an instance.

    Attributes
    ----------
    name : str
        Series name.
    primekeys : List[str]
        Series primekeys.
    keywords : pandas.DataFrame
        Details about series keywords.
    links : pandas.DataFrame
        Details about series links.
    segments : pandas.DataFrame
        Details about series segments.
    note : str
        Series description.
    dbindex : List[str]
        Series database index.
    retention : int
        Default retention time.
    unitsize : int
        Storage unit size.
    archive : int
        Series archive flag.
    tapegroup : int
        Tape group.
    """

    def __init__(self, d, *, name=None):
        self._d = d
        self.name = name
        self.retention = self._d.get("retention")
        self.unitsize = self._d.get("unitsize")
        self.archive = self._d.get("archive")
        self.tapegroup = self._d.get("tapegroup")
        self.note = self._d.get("note")
        self.primekeys = self._d.get("primekeys")
        self.dbindex = self._d.get("dbindex")
        self.keywords = self._parse_keywords(d["keywords"])
        self.links = self._parse_links(d["links"])
        self.segments = self._parse_segments(d["segments"])

    @staticmethod
    def _parse_keywords(d):
        keys = ["name", "type", "recscope", "defval", "units", "note", "linkinfo"]
        res = [tuple(di.get(k) for k in keys) for di in d]
        if not res:
            res = None  # workaround for older pandas versions
        res = pd.DataFrame(res, columns=keys)
        res.index = res.pop("name")
        res["is_time"] = res.type == "time"
        res["is_integer"] = res.type == "short"
        res["is_integer"] |= res.type == "int"
        res["is_integer"] |= res.type == "longlong"
        res["is_real"] = res.type == "float"
        res["is_real"] |= res.type == "double"
        res["is_numeric"] = res.is_integer | res.is_real
        return res

    @staticmethod
    def _parse_links(d):
        keys = ["name", "target", "kind", "note"]
        res = [tuple(di.get(k) for k in keys) for di in d]
        if not res:
            res = None  # workaround for older pandas versions
        res = pd.DataFrame(res, columns=keys)
        res.index = res.pop("name")
        return res

    @staticmethod
    def _parse_segments(d):
        keys = ["name", "type", "units", "protocol", "dims", "note"]
        res = [tuple(di.get(k) for k in keys) for di in d]
        if not res:
            res = None  # workaround for older pandas versions
        res = pd.DataFrame(res, columns=keys)
        res.index = res.pop("name")
        return res

    def __repr__(self):
        if self.name is None:
            return "<SeriesInfo>"
        return f"<SeriesInfo: {self.name}>"


class ExportRequest:
    """
    Class for handling data export requests.

    Use :func:`Client.export` or :func:`Client.export_from_id` to create
    an instance.
    """

    _status_code_ok = 0
    _status_code_notfound = 6
    _status_codes_pending = (1, 2, _status_code_notfound)
    _status_codes_ok_or_pending = (_status_code_ok, *_status_codes_pending)

    def __init__(self, d, client):
        self._client = client
        self._requestid = None
        self._status = None
        self._download_urls_cache = None
        self._update_status(d=d)

    @classmethod
    def _create_from_id(cls, requestid, client):
        d = client._json.exp_status(requestid)
        return cls(d, client)

    def __repr__(self):
        idstr = str(None) if self._requestid is None else (f"{self._requestid}")
        return f"<ExportRequest: id={idstr}, status={int(self._status)}>"

    @staticmethod
    def _parse_data(d):
        keys = ["record", "filename"]
        res = None if d is None else [(di.get(keys[0]), di.get(keys[1])) for di in d]
        return pd.DataFrame(res, columns=keys)

    def _update_status(self, *, d=None):
        if d is None and self._requestid is not None:
            d = self._client._json.exp_status(self._requestid)
        self._d = d
        self._d_time = time.time()
        self._status = int(self._d.get("status", self._status))
        self._requestid = self._d.get("requestid", self._requestid)
        if self._requestid is None:
            # Apparently 'reqid' is used instead of 'requestid' for certain
            # protocols like 'mpg'
            self._requestid = self._d.get("reqid")
        if self._requestid == "":
            # Use None if the requestid is empty (url_quick + as-is)
            self._requestid = None

    def _raise_on_error(self, *, notfound_ok=True):
        if self._status in self._status_codes_ok_or_pending:
            if self._status != self._status_code_notfound or notfound_ok:
                return  # request has not failed (yet)
        msg = self._d.get("error")
        if msg is None:
            msg = "DRMS export request failed."
        msg += f" [status={int(self._status)}]"
        raise DrmsExportError(msg)

    def _generate_download_urls(self):
        """
        Generate download URLs for the current request.
        """
        res = self.data.copy()
        data_dir = self.dir

        # Clear first record name for movies, as it is not a DRMS record.
        if self.protocol in ["mpg", "mp4"]:
            if res.record[0].startswith("movie"):
                res.record[0] = None

        # tar exports provide only a single TAR file with full path
        if self.tarfile is not None:
            data_dir = None
            res = pd.DataFrame([(None, self.tarfile)], columns=["record", "filename"])

        # If data_dir is None, the filename column should contain the full
        # path of the file and we need to extract the basename part. If
        # data_dir contains a directory, the filename column should contain
        # only the basename and we need to join it with the directory.
        if data_dir is None:
            res = res.rename(columns={"filename": "fpath"})
            split_fpath = res.fpath.str.split("/")
            res["filename"] = [sfp[-1] for sfp in split_fpath]
        else:
            res["fpath"] = [f"{data_dir}/{filename}" for filename in res.filename]

        if self.method.startswith("url"):
            baseurl = self._client._server.http_download_baseurl
        else:
            raise RuntimeError(f"Download is not supported for export method {self.method}")

        # Generate download URLs.
        urls = []
        for fp in res.fpath:
            while fp.startswith("/"):
                fp = fp[1:]
            urls.append(urljoin(baseurl, fp))
        res["url"] = urls

        # Remove rows with missing files.
        res = res[res.filename != "NoDataFile"]
        del res["fpath"]
        return res

    @staticmethod
    def _next_available_filename(fname):
        """
        Find next available filename, append a number if necessary.
        """
        i = 1
        new_fname = fname
        while Path(new_fname).exists():
            new_fname = f"{fname}.{int(i)}"
            i += 1
        return new_fname

    @property
    def id(self):
        """
        (string) Request ID.
        """
        return self._requestid

    @property
    def status(self):
        """
        (int) Export request status.
        """
        return self._status

    @property
    def method(self):
        """
        (string) Export method.
        """
        return self._d.get("method")

    @property
    def protocol(self):
        """
        (string) Export protocol.
        """
        return self._d.get("protocol")

    @property
    def dir(self):
        """
        (string) Common directory of the requested files on the server.
        """
        if self.has_finished(skip_update=True):
            self._raise_on_error()
        else:
            self.wait()
        data_dir = self._d.get("dir")
        return data_dir if data_dir else None

    @property
    def data(self):
        """
        (pandas.DataFrame) Records and filenames of the export request.

        Returns a pandas.DataFrame containing the records and filenames
        of the export request (DataFrame columns: 'record', 'filename').
        """
        if self.has_finished(skip_update=True):
            self._raise_on_error()
        else:
            self.wait()
        return self._parse_data(self._d.get("data"))

    @property
    def tarfile(self):
        """
        (string) Filename, if a TAR file was requested.
        """
        if self.has_finished(skip_update=True):
            self._raise_on_error()
        else:
            self.wait()
        data_tarfile = self._d.get("tarfile")
        return data_tarfile if data_tarfile else None

    @property
    def keywords(self):
        """
        (string) Filename of textfile containing record keywords.
        """
        if self.has_finished(skip_update=True):
            self._raise_on_error()
        else:
            self.wait()
        data_keywords = self._d.get("keywords")
        return data_keywords if data_keywords else None

    @property
    def request_url(self):
        """
        (string) URL of the export request.
        """
        data_dir = self.dir
        http_baseurl = self._client._server.http_download_baseurl
        if data_dir is None or http_baseurl is None:
            return None
        if data_dir.startswith("/"):
            data_dir = data_dir[1:]
        return urljoin(http_baseurl, data_dir)

    @property
    def urls(self):
        """
        (pandas.DataFrame) URLs of all downloadable files.

        Returns a pandas.DataFrame containing the records, filenames and
        URLs of the export request (DataFrame columns: 'record',
        'filename' and 'url').
        """
        if self._download_urls_cache is None:
            self._download_urls_cache = self._generate_download_urls()
        return self._download_urls_cache

    def has_finished(self, *, skip_update=False):
        """
        Check if the export request has finished.

        Parameters
        ----------
        skip_update : bool
            If set to True, the export status will not be updated from
            the server, even if it was in pending state after the last
            status update.

        Returns
        -------
        result : bool
            True if the export request has finished or False if the
            request is still pending.
        """
        pending = self._status in self._status_codes_pending
        if not pending:
            return True
        if not skip_update:
            self._update_status()
            pending = self._status in self._status_codes_pending
        return not pending

    def has_succeeded(self, *, skip_update=False):
        """
        Check if the export request has finished successfully.

        Parameters
        ----------
        skip_update : bool
            If set to True, the export status will not be updated from
            the server, even if it was in pending state after the last
            status update.

        Returns
        -------
        result : bool
            True if the export request has finished successfully or
            False if the request failed or is still pending.
        """
        if not self.has_finished(skip_update=skip_update):
            return False
        return self._status == self._status_code_ok

    def has_failed(self, *, skip_update=False):
        """
        Check if the export request has finished unsuccessfully.

        Parameters
        ----------
        skip_update : bool
            If set to True, the export status will not be updated from
            the server, even if it was in pending state after the last
            status update.

        Returns
        -------
        result : bool
            True if the export request has finished unsuccessfully or
            False if the request has succeeded or is still pending.
        """
        if not self.has_finished(skip_update=skip_update):
            return False
        return self._status not in self._status_codes_ok_or_pending

    def wait(self, *, timeout=None, sleep=5, retries_notfound=5):
        """
        Wait for the server to process the export request. This method
        continuously updates the request status until the server signals that
        the export request has succeeded or failed.

        Parameters
        ----------
        timeout : int or None
            Maximum number of seconds until this method times out. If
            set to `None` (the default), the status will be updated
            indefinitely until the request succeeded or failed.
        sleep : int or None
            Time in seconds between status updates (defaults to 5
            seconds). If set to None, a server supplied value is used.
        retries_notfound : int
            Number of retries in case the request was not found on the
            server. Note that it usually takes a short time until a new
            request is registered on the server, so a value too low
            might cause an exception to be raised, even if the request
            is valid and will eventually show up on the server.

        Returns
        -------
        result : bool
            True if the request succeeded or False if a timeout
            occurred. In case of an error an exception is raised.
        """
        if timeout is not None:
            t_start = time.time()
            timeout = float(timeout)
        if sleep is not None:
            sleep = float(sleep)
        retries_notfound = int(retries_notfound)

        # We are done, if the request has already finished.
        if self.has_finished(skip_update=True):
            self._raise_on_error()
            return True

        while True:
            idstr = str(None) if self._requestid is None else (f"{self._requestid}")
            logger.info(f"Export request pending. [id={idstr}, status={self._status}]")

            # Use the user-provided sleep value or the server's wait value.
            # In case neither is available, wait for 5 seconds.
            wait_secs = self._d.get("wait", 5) if sleep is None else sleep

            # Consider the time that passed since the last status update.
            wait_secs -= time.time() - self._d_time
            if wait_secs < 0:
                wait_secs = 0

            if timeout is not None:
                # Return, if we would time out while sleeping.
                if t_start + timeout + wait_secs - time.time() < 0:
                    return False

            logger.info(f"Waiting for {int(round(wait_secs))} seconds...")
            time.sleep(wait_secs)

            if self.has_finished():
                self._raise_on_error()
                return True
            if self._status == self._status_code_notfound:
                # Raise exception, if no retries are left.
                if retries_notfound <= 0:
                    self._raise_on_error(notfound_ok=False)
                logger.info(f"Request not found on server, {retries_notfound} retries left.")
                retries_notfound -= 1

    def download(self, directory, *, index=None, fname_from_rec=None, timeout=60):
        """
        Download data files.

        By default, the server-side filenames are used as local
        filenames, except for export method 'url_quick', where the
        local filenames are generated from record names (see parameter
        fname_from_rec). In case a file with the same name already
        exists in the download directory, an ascending number is
        appended to the filename.

        Note: Downloading data segments that are directories, e.g. data
        segments from series like "hmi.rdVflows_fd15_frame", is
        currently not supported. In order to download data from series
        like this, you need to use the export method 'url-tar'
        when submitting the data export request.

        Parameters
        ----------
        directory : str
            Download directory (must already exist).
        index : int, List[int], None
            Index (or indices) of the file(s) to be downloaded. If set
            to None (the default), all files of the export request are
            downloaded. Note that this parameter is ignored for export
            method 'url-tar', where only a single tar file is available
            for download.
        fname_from_rec : bool or None
            If True, local filenames are generated from record names.
            If set to False, the original filenames are used. If set to
            None (default), local filenames are generated only for
            export method 'url_quick'. Exceptions: For exports with
            method 'url-tar', no filename will be
            generated. This also applies to movie files from exports
            with protocols 'mpg' or 'mp4', where the original filename
            is used locally.
        timeout: float, optional
            Sets the timeout to "urlopen", this defaults to 60 seconds.
            This can be overridden if you set the socket timeout using
            `socket.setdefaulttimeout`.

        Returns
        -------
        result : pandas.DataFrame
            DataFrame containing the record string, download URL and
            local location of each downloaded file (DataFrame columns:
            'record', 'url' and 'download').
        """
        out_dir = Path(directory).absolute()
        if not out_dir.is_dir():
            raise OSError(f"Download directory {out_dir} does not exist")

        if np.isscalar(index):
            index = [int(index)]
        elif index is not None:
            index = list(index)

        # Wait until the export request has finished.
        self.wait()

        if fname_from_rec is None:
            # For 'url_quick', generate local filenames from record strings.
            if self.method == "url_quick":
                fname_from_rec = True

        # self.urls contains the same records as self.data, except for the tar
        # methods, where self.urls only contains one entry, the TAR file.
        data = self.urls
        if index is not None and self.tarfile is None:
            data = data.iloc[index].copy()
        ndata = len(data)

        downloads = []
        for i in range(ndata):
            di = data.iloc[i]
            if fname_from_rec:
                filename = self._client._filename_from_export_record(di.record, old_fname=di.filename)
                if filename is None:
                    filename = di.filename
            else:
                filename = di.filename

            fpath = Path(out_dir) / filename
            fpath_new = self._next_available_filename(fpath)
            fpath_tmp = self._next_available_filename(f"{fpath_new}.part")
            logger.info(f"Downloading file {int(i + 1)} of {int(ndata)}...")
            logger.info(f"    record: {di.record}")
            logger.info(f"    filename: {di.filename}")
            try:
                timeout = socket.getdefaulttimeout() or timeout
                with (
                    urlopen(create_request_with_header(di.url), timeout=timeout) as response,
                    open(fpath_tmp, "wb") as out_file,
                ):
                    shutil.copyfileobj(response, out_file)
            except (HTTPError, URLError):
                fpath_new = None
                logger.info("    -> Error: Could not download file")
            else:
                fpath_new = self._next_available_filename(fpath)
                Path(fpath_tmp).rename(fpath_new)
                logger.info(f"    -> {os.path.relpath(fpath_new)}")
            downloads.append(fpath_new)

        res = data[["record", "url"]].copy()
        res["download"] = downloads
        return res


class Client:
    """
    Client for remote DRMS server access.

    Parameters
    ----------
    server : str or drms.config.ServerConfig
        Registered server ID or ServerConfig instance.
        Defaults to JSOC.
    email : str or None
        Default email address used data export requests.
    """

    def __init__(self, server="jsoc", *, email=None):
        self._json = HttpJsonClient(server)
        self._info_cache = {}
        self.email = email  # use property for email validation

    def __repr__(self):
        return f"<Client: {self._server.name}>"

    def _convert_numeric_keywords(self, ds, kdf, *, skip_conversion=None):
        si = self.info(ds)
        int_keys = list(si.keywords[si.keywords.is_integer].index)
        num_keys = list(si.keywords[si.keywords.is_numeric].index)
        num_keys += ["*recnum*", "*sunum*", "*size*"]
        if skip_conversion is None:
            skip_conversion = []
        elif isinstance(skip_conversion, str):
            skip_conversion = [skip_conversion]
        for k in kdf:
            if k in skip_conversion:
                continue
            # pandas apparently does not support hexadecimal strings, so
            # we need a special treatment for integer strings that start
            # with '0x', like QUALITY. The following to_numeric call is
            # still necessary as the results are still Python objects.
            if k in int_keys and kdf[k].dtype is np.dtype(object):
                idx = kdf[k].str.startswith("0x")
                if idx.any():
                    k_idx = kdf.columns.get_loc(k)
                    kdf.loc[idx, kdf.columns[k_idx]] = kdf.loc[idx, kdf.columns[k_idx]].apply(int, base=16)
            if k in num_keys:
                kdf[k] = _pd_to_numeric_coerce(kdf[k])

    @staticmethod
    def _raise_query_error(d, *, status=None):
        """
        Raises a DrmsQueryError, using the json error message from d.
        """
        if status is None:
            status = d.get("status")
        msg = d.get("error")
        if msg is None:
            msg = "DRMS Query failed."
        msg += f" [status={status}]"
        raise DrmsQueryError(msg)

    def _generate_filenamefmt(self, sname):
        """
        Generate filename format string for export requests.
        """
        try:
            si = self.info(sname)
        except Exception as e:  # NOQA: BLE001
            # Cannot generate filename format for unknown series.
            logger.warning(f"Cannot generate filename format for unknown series '{sname}' with {e}")
            return None

        pkfmt_list = []
        for k in si.primekeys:
            if si.keywords.loc[k].is_time:
                pkfmt_list.append(f"{{{k}:A}}")
            else:
                pkfmt_list.append(f"{{{k}}}")

        if pkfmt_list:
            return "{}.{}.{{segment}}".format(si.name, ".".join(pkfmt_list))
        return str(si.name) + ".{recnum:%lld}.{segment}"

    # Some regular expressions used to parse export request queries.
    _re_export_recset = re.compile(r"^\s*([\w\.]+)\s*(\[.*\])?\s*(?:\{([\w\s\.,]*)\})?\s*$")
    _re_export_recset_pkeys = re.compile(r"\[([^\[^\]]*)\]")
    _re_export_recset_slist = re.compile(r"[\s,]+")

    @staticmethod
    def _parse_export_recset(rs):
        """
        Parse export request record set.
        """
        if rs is None:
            return None, None, None
        m = Client._re_export_recset.match(rs)
        if not m:
            return None, None, None
        sname, pkeys, segs = m.groups()
        if pkeys is not None:
            pkeys = Client._re_export_recset_pkeys.findall(pkeys)
        if segs is not None:
            segs = Client._re_export_recset_slist.split(segs)
        return sname, pkeys, segs

    def _filename_from_export_record(self, rs, *, old_fname=None):
        """
        Generate a filename from an export request record.
        """
        sname, pkeys, segs = self._parse_export_recset(rs)
        if sname is None:
            return None

        # We need to identify time primekeys and change the time strings to
        # make them suitable for filenames.
        try:
            si = self.info(sname)
        except Exception as e:  # NOQA: BLE001
            # Cannot generate filename for unknown series.
            logger.warning(f"Cannot generate filename format for unknown series '{sname}' with {e}")
            return None

        if pkeys is not None:
            n = len(pkeys)
            if n != len(si.primekeys):
                # Number of parsed pkeys differs from series definition.
                return None
            for i in range(n):
                # Cleanup time strings.
                if si.keywords.loc[si.primekeys[i]].is_time:
                    v = pkeys[i]
                    v = v.replace(".", "").replace(":", "").replace("-", "")
                    pkeys[i] = v

        # Generate filename.
        fname = si.name
        if pkeys is not None:
            pkeys = [k for k in pkeys if k.strip()]
            pkeys_str = ".".join(pkeys)
            if pkeys_str:
                fname += f".{pkeys_str}"
        if segs is not None:
            segs = [s for s in segs if s.strip()]
            segs_str = ".".join(segs)
            if segs_str:
                fname += f".{segs_str}"

        if old_fname is not None:
            # Try to use the file extension of the original filename.
            known_fname_extensions = [".fits", ".txt", ".jpg", ".mpg", ".mp4", ".tar"]
            for ext in known_fname_extensions:
                if old_fname.endswith(ext):
                    return fname + ext
        return fname

    # Export color table names, from (internal) series "jsoc.Color_Tables"
    _export_color_table_names = (
        "HMI_mag.lut",
        "aia_131.lut",
        "aia_1600.lut",
        "aia_1700.lut",
        "aia_171.lut",
        "aia_193.lut",
        "aia_211.lut",
        "aia_304.lut",
        "aia_335.lut",
        "aia_4500.lut",
        "aia_94.lut",
        "aia_mixed",
        "bb.sao",
        "grey.sao",
        "heat.sao",
    )

    # Export scaling types, from (internal) series "jsoc.Color_Tables"
    _export_scaling_names = ("LOG", "MINMAX", "MINMAXGIVEN", "SQRT", "mag")

    @staticmethod
    def _validate_export_protocol_args(protocol_args):
        """
        Validate export protocol arguments.
        """
        if protocol_args is None:
            return

        ct_key = "ct"
        ct = protocol_args.get(ct_key)
        if ct is None:
            ct_key = "CT"
            ct = protocol_args.get(ct_key)
        if ct is not None:
            ll = [s.lower() for s in Client._export_color_table_names]
            try:
                i = ll.index(ct.lower())
            except ValueError as e:
                msg = f"{ct} is not a valid color table, "
                msg += "available color tables: {}".format(
                    ", ".join([str(s) for s in Client._export_color_table_names]),
                )
                raise ValueError(msg) from e
            protocol_args[ct_key] = Client._export_color_table_names[i]

        scaling = protocol_args.get("scaling")
        if scaling is not None:
            ll = [s.lower() for s in Client._export_scaling_names]
            try:
                i = ll.index(scaling.lower())
            except ValueError as e:
                msg = f"{scaling} is not a valid scaling type,"
                msg += "available scaling types: {}".format(", ".join([str(s) for s in Client._export_scaling_names]))
                raise ValueError(msg) from e
            protocol_args["scaling"] = Client._export_scaling_names[i]

    @property
    def _server(self):
        """
        (ServerConfig) Remote server configuration.
        """
        return self._json.server

    @property
    def email(self):
        """
        (string) Default email address used for data export requests.
        """
        return self._email

    @email.setter
    def email(self, value):
        if value is not None and not self.check_email(value):
            raise ValueError("Email address is invalid or not registered")
        self._email = value

    def series(self, regex=None, *, full=False):
        """
        List available data series.

        Parameters
        ----------
        regex : str or None, optional
            Regular expression, used to select a subset of the
            available series. If set to None, a list of all available
            series is returned.
        full : bool
            If True, return a pandas.DataFrame containing additional
            series information, like description and primekeys. If
            False (default), the result is a list containing only the
            series names.

        Returns
        -------
        result : list or pandas.DataFrame
            List of series names or DataFrame containing name,
            primekeys and a description of the selected series (see
            parameter ``full``).
        """
        if not self._server.check_supported("series"):
            raise DrmsOperationNotSupported("Server does not support series list access")
        if self._server.url_show_series_wrapper is None:
            # No wrapper CGI available, use the regular version.
            d = self._json.show_series(ds_filter=regex)
            status = d.get("status")
            if status != 0:
                self._raise_query_error(d)
            if full:
                keys = ("name", "primekeys", "note")
                if not d["names"]:
                    return pd.DataFrame(columns=keys)
                recs = [(it["name"], _split_arg(it["primekeys"]), it["note"]) for it in d["names"]]
                return pd.DataFrame(recs, columns=keys)
            if not d["names"]:
                return []
            return [it["name"] for it in d["names"]]
        # Use show_series_wrapper instead of the regular version.
        d = self._json.show_series_wrapper(ds_filter=regex, info=full)
        if full:
            keys = ("name", "note")
            if not d["seriesList"]:
                return pd.DataFrame(columns=keys)
            recs = []
            for it in d["seriesList"]:
                name, info = next(iter(it.items()))
                note = info.get("description", "")
                recs.append((name, note))
            return pd.DataFrame(recs, columns=keys)
        return d["seriesList"]

    def info(self, ds):
        """
        Get information about the content of a data series.

        Parameters
        ----------
        ds : str
            Name of the data series.

        Returns
        -------
        result : `SeriesInfo`
            SeriesInfo instance containing information about the data
            series.
        """
        if not self._server.check_supported("info"):
            raise DrmsOperationNotSupported("Server does not support series info access")
        name = _extract_series_name(ds)
        if name is not None:
            name = name.lower()
        if name in self._info_cache:
            return self._info_cache[name]
        d = self._json.series_struct(name)
        status = d.get("status")
        if status != 0:
            self._raise_query_error(d)
        si = SeriesInfo(d, name=name)
        if name is not None:
            self._info_cache[name] = si
        return si

    def keys(self, ds):
        """
        Get a list of keywords that are available for a series. Use the
        :func:`info` method for more details.

        Parameters
        ----------
        ds : str
            Name of the data series.

        Returns
        -------
        result : list
            List of keywords available for the selected series.
        """
        si = self.info(ds)
        return list(si.keywords.index)

    def pkeys(self, ds):
        """
        Get a list of primekeys that are available for a series. Use the
        :func:`info` method for more details.

        Parameters
        ----------
        ds : str
            Name of the data series.

        Returns
        -------
        result : list
            List of primekeys available for the selected series.
        """
        si = self.info(ds)
        return list(si.primekeys)

    def query(
        self,
        ds,
        *,
        key=None,
        seg=None,
        link=None,
        convert_numeric=True,
        skip_conversion=None,
        pkeys=False,
        rec_index=False,
        n=None,
    ):
        """
        Query keywords, segments and/or links of a record set. At least one of
        the parameters key, seg, link or pkeys needs to be specified.

        Parameters
        ----------
        ds : str
            Record set query.
        key : str, List[str] or None
            List of requested keywords, optional. If set to None
            (default), no keyword results will be returned, except
            when pkeys is True.
        seg : str, List[str] or None
            List of requested segments, optional. If set to None
            (default), no segment results will be returned.
        link : str, List[str] or None
            List of requested Links, optional. If set to None
            (default), no link results will be returned.
        convert_numeric : bool
            Convert keywords with numeric types from string to
            numbers. This may result in NaNs for invalid/missing
            values. Default is True.
        skip_conversion : List[str] or None
            List of keywords names to be skipped when performing a
            numeric conversion. Default is None.
        pkeys : bool
            If True, all primekeys of the series are added to the
            ``key`` parameter.
        rec_index : bool
            If True, record names are used as index for the resulting
            DataFrames.
        n : int or None
            Limits the number of records returned by the query. For
            positive
            values, the first n records of the record set are
            returned, for negative values the last abs(n) records. If
            set to None (default), no limit is applied.

        Returns
        -------
        res_key : pandas.DataFrame, optional
            Keyword query results. This DataFrame is only returned,
            if key is not None or pkeys is set to True.
        res_seg : pandas.DataFrame, optional
            Segment query results. This DataFrame is only returned,
            if seg is not None.
        res_link : pandas.DataFrame, optional
            Link query results. This DataFrame is only returned,
            if link is not None.
        """
        if not self._server.check_supported("query"):
            raise DrmsOperationNotSupported("Server does not support DRMS queries")
        if pkeys:
            pk = self.pkeys(ds)
            key = _split_arg(key) if key is not None else []
            key = [k for k in key if k not in pk]
            key = pk + key

        lres = self._json.rs_list(ds, key=key, seg=seg, link=link, recinfo=rec_index, n=n)
        status = lres.get("status")
        if status != 0:
            self._raise_query_error(lres)

        res = []
        if key is not None:
            if "keywords" in lres:
                names = [it["name"] for it in lres["keywords"]]
                values = [it["values"] for it in lres["keywords"]]
                res_key = pd.DataFrame.from_dict(OrderedDict(zip(names, values, strict=False)))
            else:
                res_key = pd.DataFrame()
            if convert_numeric:
                self._convert_numeric_keywords(ds, res_key, skip_conversion=skip_conversion)
            res.append(res_key)

        if seg is not None:
            if "segments" in lres:
                names = [it["name"] for it in lres["segments"]]
                values = [it["values"] for it in lres["segments"]]
                res_seg = pd.DataFrame.from_dict(OrderedDict(zip(names, values, strict=False)))
            else:
                res_seg = pd.DataFrame()
            res.append(res_seg)

        if link is not None:
            if "links" in lres:
                names = [it["name"] for it in lres["links"]]
                values = [it["values"] for it in lres["links"]]
                res_link = pd.DataFrame.from_dict(OrderedDict(zip(names, values, strict=False)))
            else:
                res_link = pd.DataFrame()
            res.append(res_link)

        if rec_index:
            index = [it["name"] for it in lres["recinfo"]]
            for r in res:
                r.index = index

        if len(res) == 0:
            return None
        if len(res) == 1:
            return res[0]
        return tuple(res)

    def check_email(self, email):
        """
        Check if the email address is registered for data export.

        You can register your email for data exports from JSOC on the
        `JSOC email registration <http://jsoc.stanford.edu/ajax/register_email.html>`__
        webpage.

        Parameters
        ----------
        email : str
            Email address to be checked.

        Returns
        -------
        result : bool
            True if the email address is valid and registered, False
            otherwise.
        """
        if not self._server.check_supported("email"):
            raise DrmsOperationNotSupported("Server does not support user emails")
        res = self._json.check_address(email)
        status = res.get("status")
        return status is not None and int(status) == 2

    def export(
        self,
        ds,
        *,
        method="url_quick",
        protocol="as-is",
        protocol_args=None,
        filenamefmt=None,
        n=None,
        email=None,
        requester=None,
        process=None,
    ):
        """
        Submit a data export request.

        A registered email address is required for data exports. You
        can register your email address for data exports from JSOC on
        the `JSOC email registration
        <http://jsoc.stanford.edu/ajax/register_email.html>`__
        webpage.

        An interactive webinterface and additional information is
        available on the `JSOC data export
        <http://jsoc.stanford.edu/ajax/exportdata.html>`__ webpage.

        Note that export requests that were submitted using the
        webinterface can be accessed using the :func:`export_from_id`
        method.

        Parameters
        ----------
        ds : str
            Data export record set query.
        method : str
            Export method. Supported methods are: 'url_quick', 'url'
            and 'url-tar'. Default is 'url_quick'.
        protocol : str
            Export protocol. Supported protocols are: 'as-is', 'fits',
            'jpg', 'mpg' and 'mp4'. Default is 'as-is'.
        protocol_args : dict
            Extra protocol arguments for protocols 'jpg', 'mpg' and
            'mp4'. Valid arguments are: 'ct', 'scaling', 'min', 'max'
            and 'size'. See the JSOC data export webpage for more
            details.
        filenamefmt : str, None or bool
            Custom filename format string for exported files. This is
            ignored for 'url_quick'/'as-is' data exports. If set to
            None (default), the format string will be generated using
            the primekeys of the data series. If set to False, the
            filename format string will be omitted in the export
            request.
        n : int or None
            Limits the number of records requested. For positive
            values, the first n records of the record set are returned,
            for negative values the last abs(n) records. If set to None
            (default), no limit is applied.
        email : str or None
            Registered email address. If email is None (default), the
            current default email address is used, which in this case
            has to be set before calling export() by using the
            :attr:`Client.email` attribute.
        process : `dict`, None
            Dictionary of processing commands. Each entry is also a `dict`
            containing all of the applicable options for that processing
            command. Note that only the name of the process, and not the
            arguments, are validated by the `~drms.client.Client`. In the case of invalid
            or malformed processing arguments, JSOC may still return
            an unprocessed image without the export request failing.
        requester : str, None or bool
            Export user ID. Default is None, in which case the user
            name is determined from the email address. If set to False,
            the requester argument will be omitted in the export
            request.

        Returns
        -------
        result : `ExportRequest`
        """
        if not self._server.check_supported("export"):
            raise DrmsOperationNotSupported("Server does not support export requests")
        if email is None:
            if self._email is None:
                raise ValueError("The email argument is required, when no default email address was set.")
            email = self._email

        if filenamefmt is None:
            sname = _extract_series_name(ds)
            filenamefmt = self._generate_filenamefmt(sname)
        elif filenamefmt is False:
            filenamefmt = None

        if protocol.lower() in ["jpg", "mpg", "mp4"]:
            self._validate_export_protocol_args(protocol_args)

        d = self._json.exp_request(
            ds,
            email,
            method=method,
            protocol=protocol,
            protocol_args=protocol_args,
            filenamefmt=filenamefmt,
            n=n,
            requester=requester,
            process=process,
        )
        return ExportRequest(d, client=self)

    def export_from_id(self, requestid):
        """
        Create an `ExportRequest` instance from an existing requestid.

        Parameters
        ----------
        requestid : str
            Export request ID.

        Returns
        -------
        result : `ExportRequest`
        """
        if not self._server.check_supported("export"):
            raise DrmsOperationNotSupported("Server does not support export requests")
        return ExportRequest._create_from_id(requestid, client=self)



"""
==============================================
Print all HMI series and prime keys with notes
==============================================

This example shows how to find and display the HMI series names, prime keys and corresponding notes.
"""

import textwrap

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.
client = drms.Client()

###############################################################################
# Get all available HMI series and print their names, prime keys and notes.

hmi_series = client.series(regex=r"hmi\.", full=True)

# Print series names, prime-keys (pkeys) and notes
for series in hmi_series.index:
    print("Series:", hmi_series.name[series])
    print(" Notes:", (f"\n{8 * ' '}").join(textwrap.wrap(hmi_series.note[series])))



from unittest.mock import patch

import pytest

from drms.client import Client
from drms.json import HttpJsonRequest, JsocInfoConstants


@pytest.mark.remote_data()
def test_jsocinfoconstants():
    assert isinstance(JsocInfoConstants.all, str)
    assert JsocInfoConstants.all == "**ALL**"
    client = Client()
    client.query("hmi.synoptic_mr_720s[2150]", key=JsocInfoConstants.all, seg="synopMr")


def test_request_headers():
    with patch("drms.json.urlopen") as mock:
        HttpJsonRequest("http://example.com", "latin1")

    actual_request = mock.call_args[0][0]
    assert actual_request.headers["User-agent"]
    assert "drms/" in actual_request.headers["User-agent"]
    assert "python/" in actual_request.headers["User-agent"]
    assert actual_request.full_url == "http://example.com"



import pytest

import drms


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize("method", ["url_quick", "url"])
def test_export_asis_basic(jsoc_client_export, method):
    r = jsoc_client_export.export(
        "hmi.v_sht_2drls[2024.09.19_00:00:00_TAI]{split,rot,err}",
        protocol="as-is",
        method=method,
        requester=False,
    )

    assert isinstance(r, drms.ExportRequest)
    assert r.wait(timeout=60)
    assert r.has_succeeded()
    assert r.protocol == "as-is"
    assert len(r.urls) == 9  # 3 files per segment

    for record in r.urls.record:
        record = record.lower()
        assert record.startswith("hmi.v_sht_2drls[2024.09.19_00:00:00_tai]")
        assert record.endswith(("{split}", "{rot}", "{err}"))

    for filename in r.urls.filename:
        assert filename.endswith(("err.2d", "rot.2d", "splittings.out"))

    for url in r.urls.url:
        assert url.endswith(("err.2d", "rot.2d", "splittings.out"))


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_export_fits_basic(jsoc_client_export):
    r = jsoc_client_export.export(
        "hmi.sharp_720s[4864][2014.11.30_00:00_TAI]{continuum, magnetogram}",
        protocol="fits",
        method="url",
        requester=False,
    )

    assert isinstance(r, drms.ExportRequest)
    assert r.wait(timeout=60)
    assert r.has_succeeded()
    assert r.protocol == "fits"
    assert len(r.urls) == 2  # 1 file per segment

    for record in r.urls.record:
        record = record.lower()
        assert record.startswith("hmi.sharp_720s[4864]")
        assert record.endswith("2014.11.30_00:00:00_tai]")

    for filename in r.urls.filename:
        assert filename.endswith(("continuum.fits", "magnetogram.fits"))

    for url in r.urls.url:
        assert url.endswith(("continuum.fits", "magnetogram.fits"))


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_export_im_patch(jsoc_client_export):
    # TODO: check that this has actually done the export/processing properly?
    # NOTE: processing exports seem to fail silently on the server side if
    # the correct names/arguments are not passed. Not clear how to check
    # that this has not happened.
    process = {
        "im_patch": {
            "t_ref": "2025-01-01T04:33:30.000",
            "t": 0,
            "r": 0,
            "c": 0,
            "locunits": "arcsec",
            "boxunits": "arcsec",
            "x": -517.2,
            "y": -246,
            "width": 345.6,
            "height": 345.6,
        },
    }
    req = jsoc_client_export.export(
        "aia.lev1_euv_12s[2025-01-01T04:33:30.000/1m@12s][171]{image}",
        method="url",
        protocol="fits",
        process=process,
        requester=False,
    )

    assert isinstance(req, drms.ExportRequest)
    assert req.wait(timeout=60)
    assert req.has_succeeded()
    assert req.protocol == "fits"

    for record in req.urls.record:
        record = record.lower()
        assert record.startswith("aia.lev1_euv_12s_mod")

    for filename in req.urls.filename:
        assert filename.endswith("image.fits")

    for url in req.urls.url:
        assert url.endswith("image.fits")


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_export_rebin(jsoc_client_export):
    # TODO: check that this has actually done the export/processing properly?
    # NOTE: processing exports seem to fail silently on the server side if
    # the correct names/arguments are not passed. Not clear how to check
    # that this has not happened.
    req = jsoc_client_export.export(
        "hmi.M_720s[2020-10-17_22:12:00_TAI/24m]{magnetogram}",
        method="url",
        protocol="fits",
        process={"rebin": {"method": "boxcar", "scale": 0.25}},
        requester=False,
    )

    assert isinstance(req, drms.ExportRequest)
    assert req.wait(timeout=60)
    assert req.has_succeeded()
    assert req.protocol == "fits"

    for record in req.urls.record:
        record = record.lower()
        assert record.startswith("hmi.m_720s_mod")

    for filename in req.urls.filename:
        assert filename.endswith("magnetogram.fits")

    for url in req.urls.url:
        assert url.endswith("magnetogram.fits")


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_export_invalid_process(jsoc_client_export):
    with pytest.raises(ValueError, match="foobar is not one of the allowed processing options"):
        jsoc_client_export.export(
            "aia.lev1_euv_12s[2015-10-17T04:33:30.000/1m@12s][171]{image}",
            process={"foobar": {}},
        )


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_export_email(jsoc_client):
    with pytest.raises(ValueError, match="The email argument is required, when no default email address was set."):
        jsoc_client.export("hmi.v_45s[2016.04.01_TAI/1d@6h]{Dopplergram}")



"""
==================================
Exporting from existing RequestIDs
==================================

This example takes a RequestID of an already existing export request, prints
the corresponding "Request URL" and downloads the available files.

Note that you can also use RequestIDs from export requests, that were
submitted using the JSOC website.
"""

from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# Export request ID
request_id = "JSOC_20201101_198"

# Querying the server using the entered RequestID.
print(f"Looking up export request {request_id}...")
result = client.export_from_id(request_id)

# Print request URL and number of available files.
print(f"\nRequest URL: {result.request_url}")
print(f"{int(len(result.urls))} file(s) available for download.\n")

# Create download directory if it does not exist yet.
out_dir = Path("downloads") / request_id
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

# Download all available files.
result.download(out_dir)
print("Download finished.")
print(f"\nDownload directory:\n  {Path.resolve(out_dir)}\n")



from urllib.parse import urljoin

__all__ = ["ServerConfig", "register_server"]


class ServerConfig:
    """
    DRMS Server configuration.

    Additional keyword arguments can be used to add additional entries
    to config. In case a keyword argument already exists in the config
    dictionary, the config entry will be replaced by the kwargs value.

    Available config keys are:
        name
        cgi_baseurl
        cgi_show_series
        cgi_jsoc_info
        cgi_jsoc_fetch
        cgi_check_address
        cgi_show_series_wrapper
        show_series_wrapper_dbhost
        url_show_series
        url_jsoc_info
        url_jsoc_fetch
        url_check_address
        url_show_series_wrapper
        encoding
        http_download_baseurl

    Parameters
    ----------
    name : str
        Server configuration name.
    config : dict
        Dictionary containing configuration entries (see below for a
        list of available entries).
    """

    _valid_keys = (
        "name",
        "cgi_baseurl",
        "cgi_show_series",
        "cgi_jsoc_info",
        "cgi_jsoc_fetch",
        "cgi_check_address",
        "cgi_show_series_wrapper",
        "show_series_wrapper_dbhost",
        "url_show_series",
        "url_jsoc_info",
        "url_jsoc_fetch",
        "url_check_address",
        "url_show_series_wrapper",
        "encoding",
        "http_download_baseurl",
    )

    def __init__(self, config=None, **kwargs):
        self._d = d = config.copy() if config is not None else {}
        d.update(kwargs)

        for k in d:
            if k not in self._valid_keys:
                raise ValueError(f"Invalid server config key: {k}")

        if "name" not in d:
            raise ValueError('Server config entry "name" is missing')

        # encoding defaults to latin1
        if "encoding" not in d:
            d["encoding"] = "latin1"

        # Generate URL entries from CGI entries, if cgi_baseurl exists and
        # the specific URL entry is not already set.
        if "cgi_baseurl" in d:
            cgi_baseurl = d["cgi_baseurl"]
            cgi_keys = [k for k in self._valid_keys if k.startswith("cgi") and k != "cgi_baseurl"]
            for k in cgi_keys:
                url_key = f"url{k[3:]}"
                cgi_value = d.get(k)
                if d.get(url_key) is None and cgi_value is not None:
                    d[url_key] = urljoin(cgi_baseurl, cgi_value)

    def __repr__(self):
        return f"<ServerConfig: {self._d.get('name')}>"

    def __dir__(self):
        return dir(type(self)) + list(self.__dict__.keys()) + list(self._valid_keys)

    def __getattr__(self, name):
        if name in self._valid_keys:
            return self._d.get(name)
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name in self._valid_keys:
            if not isinstance(value, str):
                raise ValueError(f"{name} config value must be a string")
            self._d[name] = value
        else:
            object.__setattr__(self, name, value)

    def copy(self):
        return ServerConfig(self._d)

    def to_dict(self):
        return self._d

    def check_supported(self, op):
        """
        Check if an operation is supported by the server.
        """
        if op == "series":
            return (self.cgi_show_series is not None) or (self.cgi_show_series_wrapper is not None)
        if op == "info":
            return self.cgi_jsoc_info is not None
        if op == "query":
            return self.cgi_jsoc_info is not None
        if op == "email":
            return self.cgi_check_address is not None
        if op == "export":
            return (self.cgi_jsoc_info is not None) and (self.cgi_jsoc_fetch is not None)
        raise ValueError(f"Unknown operation: {op!r}")


def register_server(config):
    """
    Register a server configuration.
    """
    global _server_configs
    name = config.name.lower()
    if name in _server_configs:
        raise RuntimeError(f"ServerConfig {name} already registered")
    _server_configs[config.name.lower()] = config


# Registered servers
_server_configs = {}

# Register public JSOC DRMS server.
register_server(
    ServerConfig(
        name="JSOC",
        cgi_baseurl="http://jsoc.stanford.edu/cgi-bin/ajax/",
        cgi_show_series="show_series",
        cgi_jsoc_info="jsoc_info",
        cgi_jsoc_fetch="jsoc_fetch",
        cgi_check_address="checkAddress.sh",
        cgi_show_series_wrapper="showextseries",
        show_series_wrapper_dbhost="hmidb2",
        http_download_baseurl="http://jsoc.stanford.edu/",
    ),
)

# Register KIS DRMS server.
register_server(
    ServerConfig(
        name="KIS",
        cgi_baseurl="http://drms.leibniz-kis.de/cgi-bin/",
        cgi_show_series="show_series",
        cgi_jsoc_info="jsoc_info",
    ),
)



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

import os
import datetime
from pathlib import Path

from sunpy_sphinx_theme import PNG_ICON

from packaging.version import Version

# -- Project information -----------------------------------------------------

# The full version, including alpha/beta/rc tags
from drms import __version__

_version = Version(__version__)
version = release = str(_version)
# Avoid "post" appearing in version string in rendered docs
if _version.is_postrelease:
    version = release = _version.base_version
# Avoid long githashes in rendered Sphinx docs
elif _version.is_devrelease:
    version = release = f"{_version.base_version}.dev{_version.dev}"
is_development = _version.is_devrelease
is_release = not (_version.is_prerelease or _version.is_devrelease)

project = "drms"
author = "The SunPy Community"
copyright = f"{datetime.datetime.now().year}, {author}"  # noqa: A001

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "hoverxref.extension",
    "sphinx_copybutton",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_changelog",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
automodapi_toctreedirnm = "generated/api"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Treat everything in single ` as a Python reference.
default_role = "py:obj"

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/python3.inv"),
    ),
    "numpy": (
        "https://numpy.org/doc/stable/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/numpy.inv"),
    ),
    "scipy": (
        "https://docs.scipy.org/doc/scipy/reference/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/scipy.inv"),
    ),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "sunpy": ("https://docs.sunpy.org/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sunpy"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# By default, when rendering docstrings for classes, sphinx.ext.autodoc will
# make docs with the class-level docstring and the class-method docstrings,
# but not the __init__ docstring, which often contains the parameters to
# class constructors across the scientific Python ecosystem. The option below
# will append the __init__ docstring to the class-level docstring when rendering
# the docs. For more options, see:
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autoclass_content
autoclass_content = "both"

# -- Other options ----------------------------------------------------------

# JSOC email os env
# see https://github.com/sunpy/sunpy/wiki/Home:-JSOC
os.environ["JSOC_EMAIL"] = "jsoc@sunpy.org"

# -- Options for hoverxref -----------------------------------------------------

if os.environ.get("READTHEDOCS"):
    hoverxref_api_host = "https://readthedocs.org"
    if os.environ.get("PROXIED_API_ENDPOINT"):
        # Use the proxied API endpoint
        # - A RTD thing to avoid a CSRF block when docs are using a
        #   custom domain
        hoverxref_api_host = "/_"

hoverxref_tooltip_maxwidth = 600  # RTD main window is 696px
hoverxref_auto_ref = True
hoverxref_mathjax = True
hoverxref_domains = ["py"]
hoverxref_role_types = {
    # roles with py domain
    "attr": "tooltip",
    "class": "tooltip",
    "const": "tooltip",
    "data": "tooltip",
    "exc": "tooltip",
    "func": "tooltip",
    "meth": "tooltip",
    "mod": "tooltip",
    "obj": "tooltip",
    # roles with std domain
    "confval": "tooltip",
    "hoverxref": "tooltip",
    "ref": "tooltip",
    "term": "tooltip",
}

# -- Options for sphinx-copybutton ---------------------------------------------

# Python Repl + continuation, Bash, ipython and qtconsole + continuation, jupyter-console + continuation
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Sphinx Gallery ------------------------------------------------------------

sphinx_gallery_conf = {
    "backreferences_dir": Path("generated") / "modules",
    "filename_pattern": "^((?!skip_).)*$",
    "examples_dirs": Path("..") / "examples",
    "within_subsection_order": "ExampleTitleSortKey",
    "gallery_dirs": Path("generated") / "gallery",
    "default_thumb_file": PNG_ICON,
    "abort_on_example_error": False,
    "plot_gallery": "True",
    "remove_config_comments": True,
    "doc_module": ("sunpy"),
    "only_warn_on_example_error": True,
    "matplotlib_animations": True,
}



import pytest

import drms
from drms.config import ServerConfig
from drms.exceptions import DrmsOperationNotSupported, DrmsQueryError


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_basic(jsoc_client):
    keys, segs = jsoc_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", key="T_REC, CRLT_OBS", seg="Dopplergram")
    assert len(keys) == 4
    for k in ["T_REC", "CRLT_OBS"]:
        assert k in keys.columns
    assert len(segs) == 4
    assert "Dopplergram" in segs.columns
    assert ((keys.CRLT_OBS - 3.14159).abs() < 0.0001).all()


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_allmissing(jsoc_client):
    with pytest.raises(ValueError, match="At least one key, seg or link must be specified"):
        jsoc_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]")


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_key(jsoc_client):
    keys = jsoc_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", key="T_REC, CRLT_OBS")
    assert len(keys) == 4
    for k in ["T_REC", "CRLT_OBS"]:
        assert k in keys.columns
    assert ((keys.CRLT_OBS - 3.14159).abs() < 0.0001).all()


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_seg(jsoc_client):
    segs = jsoc_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", seg="Dopplergram")
    assert len(segs) == 4
    assert "Dopplergram" in segs.columns


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_link(jsoc_client):
    links = jsoc_client.query("hmi.B_720s[2013.07.03_08:42_TAI/40m]", link="MDATA")
    assert len(links) == 3
    assert "MDATA" in links.columns


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_seg_key_link(jsoc_client):
    keys, segs, links = jsoc_client.query("hmi.B_720s[2013.07.03_08:42_TAI/40m]", key="foo", link="bar", seg="baz")
    assert len(keys) == 3
    assert (keys.foo == "Invalid KeyLink").all()
    assert len(segs) == 3
    assert (segs.baz == "InvalidSegName").all()
    assert len(links) == 3
    assert (links.bar == "Invalid_Link").all()


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_pkeys(jsoc_client):
    keys = jsoc_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", pkeys=True)
    pkeys = list(keys.columns.values)
    assert pkeys == jsoc_client.pkeys("hmi.v_45s[2013.07.03_08:42_TAI/3m]")
    assert len(keys) == 4


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_recindex(jsoc_client):
    keys = jsoc_client.query("hmi.V_45s[2013.07.03_08:42_TAI/4m]", key="T_REC", rec_index=True)
    index_values = list(keys.index.values)
    assert all("hmi.V_45s" in s for s in index_values)
    assert len(keys) == 5


def test_query_invalid_server():
    cfg = ServerConfig(name="TEST")
    c = drms.Client(cfg)
    with pytest.raises(DrmsOperationNotSupported):
        c.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", pkeys=True)


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_query_invalid_series(jsoc_client):
    with pytest.raises(DrmsQueryError):
        jsoc_client.query("foo", key="T_REC")


@pytest.mark.remote_data()
@pytest.mark.parametrize(
    "query",
    [
        "hmi.v_45s[2014.01.01_00:00:35_TAI-2014.01.01_01:00:35_TAI]",
        "hmi.M_720s[2011.04.14_00:30:00_TAI/6h@2h]",
    ],
)
def test_query_hexadecimal_strings(query):
    # Exercise the part of client.py that deals with hexadecimal strings
    c = drms.Client()
    c.query(query, key="**ALL**")



import pytest

import drms


@pytest.mark.kis()
@pytest.mark.remote_data()
def test_series_list_all(kis_client):
    slist = kis_client.series()
    assert isinstance(slist, list)
    assert "hmi.v_45s" in (s.lower() for s in slist)
    assert "hmi.m_720s" in (s.lower() for s in slist)
    assert "hmi.ic_720s" in (s.lower() for s in slist)
    assert "mdi.fd_v" in (s.lower() for s in slist)


@pytest.mark.kis()
@pytest.mark.remote_data()
@pytest.mark.parametrize("schema", ["hmi", "mdi"])
def test_series_list_schemata(kis_client, schema):
    regex = rf"{schema}\."
    slist = kis_client.series(regex=regex)
    assert len(slist) > 0
    for sname in slist:
        assert sname.startswith(f"{schema}.")


@pytest.mark.kis()
@pytest.mark.remote_data()
@pytest.mark.parametrize(
    ("series", "pkeys", "segments"),
    [
        ("hmi.v_45s", ["T_REC", "CAMERA"], ["Dopplergram"]),
        ("hmi.m_720s", ["T_REC", "CAMERA"], ["magnetogram"]),
        ("hmi.v_sht_2drls", ["LMIN", "NACOEFF"], ["split", "rot", "err"]),
    ],
)
def test_series_info_basic(kis_client, series, pkeys, segments):
    si = kis_client.info(series)
    assert si.name.lower() == series
    for k in pkeys:
        assert k in si.primekeys
        assert k in si.keywords.index
    for s in segments:
        assert s in si.segments.index


@pytest.mark.kis()
@pytest.mark.remote_data()
def test_query_basic(kis_client):
    keys, segs = kis_client.query("hmi.v_45s[2013.07.03_08:42_TAI/3m]", key="T_REC, CRLT_OBS", seg="Dopplergram")
    assert len(keys) == 4
    for k in ["T_REC", "CRLT_OBS"]:
        assert k in keys.columns
    assert len(segs) == 4
    assert "Dopplergram" in segs.columns
    assert ((keys.CRLT_OBS - 3.14159).abs() < 0.0001).all()


@pytest.mark.kis()
@pytest.mark.remote_data()
def test_not_supported_email(kis_client):
    with pytest.raises(drms.DrmsOperationNotSupported):
        kis_client.email = "name@example.com"


@pytest.mark.kis()
@pytest.mark.remote_data()
def test_not_supported_export(kis_client):
    with pytest.raises(drms.DrmsOperationNotSupported):
        kis_client.export("hmi.v_45s[2010.05.01_TAI]")
    with pytest.raises(drms.DrmsOperationNotSupported):
        kis_client.export_from_id("KIS_20120101_123")



import os
from urllib.error import URLError, HTTPError
from urllib.request import urlopen

import pytest

from drms.utils import create_request_with_header

# Test URLs, used to check if a online site is reachable
jsoc_testurl = "http://jsoc.stanford.edu/"
kis_testurl = "http://drms.leibniz-kis.de/"


class lazily_cached:
    """
    Lazily evaluated function call with cached result.
    """

    def __init__(self, f, *args, **kwargs):
        self.func = lambda: f(*args, **kwargs)

    def __call__(self):
        if not hasattr(self, "result"):
            self.result = self.func()
        return self.result


def site_reachable(url, timeout=60):
    """
    Checks if the given URL is accessible.
    """
    try:
        urlopen(create_request_with_header(url), timeout=timeout)
    except (URLError, HTTPError):
        return False
    return True


# Create lazily evaluated, cached site checks for JSOC and KIS.
jsoc_reachable = lazily_cached(site_reachable, jsoc_testurl)
kis_reachable = lazily_cached(site_reachable, kis_testurl)


def pytest_runtest_setup(item):
    # Skip JSOC online site tests if the site is not reachable.
    if item.get_closest_marker("jsoc") is not None:
        if not jsoc_reachable():
            pytest.skip("JSOC is not reachable")

    # Skip KIS online site tests if the site is not reachable.
    if item.get_closest_marker("kis") is not None:
        if not kis_reachable():
            pytest.skip("KIS is not reachable")


@pytest.fixture()
def email():
    """
    Email address for tests.
    """
    email = os.environ.get("JSOC_EMAIL", None)
    if email is None:
        pytest.skip("No email address specified; use the JSOC_EMAIL environmental variable to enable export tests")
    return email


@pytest.fixture()
def jsoc_client():
    """
    Client fixture for JSOC online tests, does not use email.
    """
    import drms

    return drms.Client("jsoc")


@pytest.fixture()
def jsoc_client_export(email):
    """
    Client fixture for JSOC online tests, uses email if specified.
    """
    import drms

    return drms.Client("jsoc", email=email)


@pytest.fixture()
def kis_client():
    """
    Client fixture for KIS online tests.
    """
    import drms

    return drms.Client("kis")



import pytest

import drms
from drms.config import ServerConfig
from drms.exceptions import DrmsOperationNotSupported

# Invalid email addresses used for testing
invalid_emails = [
    "notregistered@example.com",
    "not-valid",
    "",
]


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize("email", invalid_emails)
def test_email_invalid_check(email):
    c = drms.Client("jsoc")
    assert not c.check_email(email)


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize("email", invalid_emails)
def test_email_invalid_set(email):
    c = drms.Client("jsoc")
    with pytest.raises(ValueError, match="Email address is invalid or not registered"):
        c.email = email


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize("email", invalid_emails)
def test_email_invalid_init(email):
    with pytest.raises(ValueError, match="Email address is invalid or not registered"):
        drms.Client("jsoc", email=email)


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_email_cmdopt_check(email):
    c = drms.Client("jsoc")
    assert c.check_email(email)


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_email_cmdopt_set(email):
    c = drms.Client("jsoc")
    c.email = email
    assert c.email == email


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_email_cmdopt_init(email):
    c = drms.Client("jsoc", email=email)
    assert c.email == email


def test_query_invalid():
    cfg = ServerConfig(name="TEST")
    with pytest.raises(DrmsOperationNotSupported):
        drms.Client(cfg, email="user@example.com")



"""
======================
Exporting data as fits
======================

This example shows how to submit a data export request using the 'fits'
protocol and how to download the requested files.

Note that the 'as-is' protocol should be used instead of 'fits', if
record keywords in the FITS headers are not needed, as it greatly
reduces the server load.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

# Use 'as-is' instead of 'fits', if record keywords are not needed in the
# FITS header. This greatly reduces the server load!
export_protocol = "fits"

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: "Series[harpnum][timespan]{data segments}"

qstr = "hmi.sharp_720s[12519][2024.12.30_22:24:00_TAI/1d@8h]{continuum, magnetogram, field}"
print(f"Data export query:\n  {qstr}\n")

# Submit export request using the 'fits' protocol
print("Submitting export request...")
result = client.export(qstr, method="url", protocol=export_protocol, email=email)

# Print request URL.
print(f"\nRequest URL: {result.request_url}")
print(f"{int(len(result.urls))} file(s) available for download.\n")

# Download selected files.
result.download(out_dir)
print("Download finished.")
print(f'\nDownload directory:\n  "{out_dir.absolute()}"\n')



"""
================================================
Downloading and plotting a HMI polar field data
================================================

This example shows how to download HMI polar field data from JSOC and make a plot.
"""

import matplotlib.pyplot as plt
import numpy as np

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Construct the DRMS query string: "Series[timespan][wavelength]"

qstr = "hmi.meanpf_720s[2010.05.01_TAI-2016.04.01_TAI@12h]"

# Send request to the DRMS server
print(f"Querying keyword data...\n -> {qstr}")
result = client.query(qstr, key=["T_REC", "CAPN2", "CAPS2"])
print(f" -> {len(result)} lines retrieved.")

# Convert T_REC strings to datetime and use it as index for the series
result.index = drms.to_datetime(result.pop("T_REC"))

# Determine smallest timestep
dt = np.diff(result.index.to_pydatetime()).min()

# Make sure the time series contains all time steps (fills gaps with NaNs)
a = result.asfreq(dt)

###############################################################################
# Plot the magnetic field values.

# Compute 30d moving average and standard deviation using a boxcar window
win_size = int(30 * 24 * 3600 / dt.total_seconds())
a_avg = a.rolling(win_size, min_periods=1, center=True).mean()
a_std = a.rolling(win_size, min_periods=1, center=True).std()

# Plot results
t = a.index.to_pydatetime()
n, mn, sn = a.CAPN2, a_avg.CAPN2, a_std.CAPN2
s, ms, ss = a.CAPS2, a_avg.CAPS2, a_std.CAPS2

# Plot smoothed data
fig, ax = plt.subplots(1, 1, figsize=(15, 7))
ax.set_title(qstr, fontsize="medium")
ax.plot(t, n, "b", alpha=0.5, label="North pole")
ax.plot(t, s, "g", alpha=0.5, label="South pole")
ax.plot(t, mn, "r", label="Moving average")
ax.plot(t, ms, "r", label="")
ax.set_xlabel("Time")
ax.set_ylabel("Mean radial field strength [G]")
ax.legend()
fig.tight_layout()

# Plot raw data
fig, ax = plt.subplots(1, 1, figsize=(15, 7))
ax.set_title(qstr, fontsize="medium")
ax.fill_between(t, mn - sn, mn + sn, edgecolor="none", facecolor="b", alpha=0.3, interpolate=True)
ax.fill_between(t, ms - ss, ms + ss, edgecolor="none", facecolor="g", alpha=0.3, interpolate=True)
ax.plot(t, mn, "b", label="North pole")
ax.plot(t, ms, "g", label="South pole")
ax.set_xlabel("Time")
ax.set_ylabel("Mean radial field strength [G]")
ax.legend()
fig.tight_layout()

plt.show()



import json as _json
import socket
from enum import Enum
from urllib.parse import urlencode, quote_plus
from urllib.request import HTTPError, urlopen

from drms import logger
from .config import ServerConfig, _server_configs
from .utils import _split_arg, create_request_with_header

__all__ = ["HttpJsonClient", "HttpJsonRequest", "JsocInfoConstants"]


# TODO: When we support 3.11, we can use StrEnum instead of Enum
class JsocInfoConstants(str, Enum):
    """
    Constants for DRMS queries.
    """

    all = "**ALL**"
    none = "**NONE**"
    recdir = "*recdir*"
    dirmtime = "*dirmtime*"
    logdir = "*logdir*"
    recnum = "*recnum*"
    sunum = "*sunum*"
    size = "*size*"
    online = "*online*"
    retain = "*retain*"
    archive = "*archive*"


class HttpJsonRequest:
    """
    Class for handling HTTP/JSON requests.

    Use `HttpJsonClient` to create an instance.
    """

    def __init__(self, url, encoding, timeout=60):
        timeout = socket.getdefaulttimeout() or timeout
        self._encoding = encoding
        try:
            self._http = urlopen(create_request_with_header(url), timeout=timeout)
        except HTTPError as e:
            e.msg = f"Failed to open URL: {e.url} with {e.code} - {e.msg}"
            raise e
        self._data_str = None
        self._data = None

    def __repr__(self):
        return f"<HttpJsonRequest: {self.url}>"

    @property
    def url(self):
        return self._http.url

    @property
    def raw_data(self):
        if self._data_str is None:
            self._data_str = self._http.read()
        return self._data_str

    @property
    def data(self):
        if self._data is None:
            self._data = _json.loads(self.raw_data.decode(self._encoding))
        return self._data


class HttpJsonClient:
    """
    HTTP/JSON communication with the DRMS server CGIs.

    Parameters
    ----------
    server : str or drms.config.ServerConfig
        Registered server ID or ServerConfig instance.
        Defaults to JSOC.
    """

    def __init__(self, server="jsoc"):
        if isinstance(server, ServerConfig):
            self._server = server
        else:
            self._server = _server_configs[server.lower()]

    def __repr__(self):
        return f"<HttpJsonClient: {self._server.name}>"

    def _json_request(self, url):
        logger.debug(f"URL for request: {url}")
        return HttpJsonRequest(url, self._server.encoding)

    @property
    def server(self):
        return self._server

    def show_series(self, ds_filter=None):
        """
        List available data series.

        Parameters
        ----------
        ds_filter : str, None, optional
            Name filter regexp.
            Default is None, which returns all available series.

        Returns
        -------
        result : dict
        """
        query = "?" if ds_filter is not None else ""
        if ds_filter is not None:
            query += urlencode({"filter": ds_filter})
        req = self._json_request(self._server.url_show_series + query)
        return req.data

    def show_series_wrapper(self, ds_filter=None, *, info=False):
        """
        List available data series.

        This is an alternative to show_series, which needs to be used
        to get a list of all available series provided by JSOC. There
        is currently no support for retrieving primekeys using this
        CGI.

        Parameters
        ----------
        ds_filter : str, None, optional
            Name filter regexp.
            Default is None, which returns all available series.
        info : bool
            If False (default), the result only contains series names.
            If set to True, the result includes a description for each
            series.

        Returns
        -------
        result : dict
        """
        query_args = {"dbhost": self._server.show_series_wrapper_dbhost}
        if ds_filter is not None:
            query_args["filter"] = ds_filter
        if info:
            query_args["info"] = "1"
        query = f"?{urlencode(query_args)}"
        req = self._json_request(self._server.url_show_series_wrapper + query)
        return req.data

    def series_struct(self, ds):
        """
        Get information about the content of a data series.

        Parameters
        ----------
        ds : str
            Name of the data series.

        Returns
        -------
        result : dict
            Dictionary containing information about the data series.
        """
        query = f"?{urlencode({'op': 'series_struct', 'ds': ds})}"
        req = self._json_request(self._server.url_jsoc_info + query)
        return req.data

    def rs_summary(self, ds):
        """
        Get summary (i.e. count) of a given record set.

        Parameters
        ----------
        ds : str
            Record set query (only one series).

        Returns
        -------
        result : dict
            Dictionary containing 'count', 'status' and 'runtime'.
        """
        query = f"?{urlencode({'op': 'rs_summary', 'ds': ds})}"
        req = self._json_request(self._server.url_jsoc_info + query)
        return req.data

    def rs_list(self, ds, *, key=None, seg=None, link=None, recinfo=False, n=None, uid=None):
        """
        Get detailed information about a record set.

        Parameters
        ----------
        ds : str
            Record set query.
        key : str, list or None
            List of requested keywords, optional.
        seg : str, list or None
            List of requested segments, optional.
        link : str or None
            List of requested Links, optional.
        recinfo : bool
            Request record info for each record in the record set.
        n : int or None
            Record set limit. For positive values, the first n records
            of the record set are returned, for negative values the
            last abs(n) records. If set to None (default), no limit is
            applied.
        uid : str or None
            Session ID used when calling rs_list CGI, optional.

        Returns
        -------
        result : dict
            Dictionary containing the requested record set information.
        """
        if key is None and seg is None and link is None:
            raise ValueError("At least one key, seg or link must be specified")
        d = {"op": "rs_list", "ds": ds}
        if key is not None:
            d["key"] = ",".join(_split_arg(key))
        if seg is not None:
            d["seg"] = ",".join(_split_arg(seg))
        if link is not None:
            d["link"] = ",".join(_split_arg(link))
        if recinfo:
            d["R"] = "1"
        if n is not None:
            d["n"] = f"{int(int(n))}"
        if uid is not None:
            d["userhandle"] = uid
        query = f"?{urlencode(d)}"
        req = self._json_request(self._server.url_jsoc_info + query)
        return req.data

    def check_address(self, email):
        """
        Check if an email address is registered for export data requests.

        Parameters
        ----------
        email : str
            Email address to be verified.

        Returns
        -------
        result : dict
            Dictionary containing 'status' and 'msg'.

            Some status codes are:
            - 2: Email address is valid and registered
            - 4: Email address has neither been validated nor registered
            - -2: Not a valid email address
        """
        query = "?" + urlencode({"address": quote_plus(email), "checkonly": "1"})
        req = self._json_request(self._server.url_check_address + query)
        return req.data

    def exp_request(self, *args, **kwargs):
        """
        Request data export.

        Parameters
        ----------
        ds : str
            Data export record set query.
        notify : str
            Registered email address.
        method : str
            Export method. Supported methods are: 'url_quick', 'url',
            and 'url-tar'. Default is 'url_quick'.
        protocol : str
            Export protocol. Supported protocols are: 'as-is', 'fits',
            'jpg', 'mpg' and 'mp4'. Default is 'as-is'.
        protocol_args : dict or None
            Extra protocol arguments for protocols 'jpg', 'mpg' and
            'mp4'. Valid arguments are: 'ct', 'scaling', 'min', 'max'
            and 'size'.
        filenamefmt : str, None
            Custom filename format string for exported files. This is
            ignored for 'url_quick'/'as-is' data exports.
        process : `dict`, None
            Dictionary of processing commands. Each entry is also a `dict`
            containing all of the applicable options for that processing
            command.
        n : int or None
            Limits the number of records requested. For positive
            values, the first n records of the record set are returned,
            for negative values the last abs(n) records. If set to None
            (default), no limit is applied.
        requester : str, None or bool
            Export user ID. Default is None, in which case the user
            name is determined from the email address. If set to False,
            the requester argument will be omitted in the export
            request.

        Returns
        -------
        result : dict
            Dictionary containing the server response to the export
            request.
        """
        req = self._json_request(self._exp_request_url(*args, **kwargs))
        return req.data

    def _exp_request_url(
        self,
        ds,
        notify,
        *,
        method="url_quick",
        protocol="as-is",
        protocol_args=None,
        filenamefmt=None,
        n=None,
        process=None,
        requester=None,
    ):
        method = method.lower()
        method_list = ["url_quick", "url", "url-tar"]
        if method not in method_list:
            raise ValueError(
                "Method {} is not supported, valid methods are: {}".format(
                    method,
                    ", ".join(str(s) for s in method_list),
                ),
            )

        protocol = protocol.lower()
        img_protocol_list = ["jpg", "mpg", "mp4"]
        protocol_list = ["as-is", "fits", *img_protocol_list]
        if protocol not in protocol_list:
            raise ValueError(
                "Protocol {} is not supported, valid protocols are: {}".format(
                    protocol,
                    ", ".join(str(s) for s in protocol_list),
                ),
            )

        # method "url_quick" is meant to be used with "as-is", change method
        # to "url" if protocol is not "as-is"
        if method == "url_quick" and protocol != "as-is":
            method = "url"

        if protocol in img_protocol_list:
            extra_keys = {"ct": "grey.sao", "scaling": "MINMAX", "size": 1}
            if protocol_args is not None:
                for k, v in protocol_args.items():
                    if k.lower() == "ct":
                        extra_keys["ct"] = v
                    elif k == "scaling":
                        extra_keys[k] = v
                    elif k == "size":
                        extra_keys[k] = int(v)
                    elif k in ["min", "max"]:
                        extra_keys[k] = float(v)
                    else:
                        raise ValueError(f"Unknown protocol argument: {k}")
            protocol += ",CT={ct},scaling={scaling},size={size}".format(**extra_keys)
            if "min" in extra_keys:
                protocol += f",min={extra_keys['min']:g}"
            if "max" in extra_keys:
                protocol += f",max={extra_keys['max']:g}"
        else:
            if protocol_args is not None:
                raise ValueError(f"protocol_args not supported for protocol {protocol}")

        d = {
            "op": "exp_request",
            "format": "json",
            "ds": ds,
            "notify": notify,
            "method": method,
            "protocol": protocol,
        }

        if filenamefmt is not None:
            d["filenamefmt"] = filenamefmt

        n = int(n) if n is not None else 0
        d["process=n"] = f"{n}"
        if process is not None:
            allowed_processes = [
                "im_patch",
                "resize",
                "rebin",
                "aia_scale_aialev1",
                "aia_scale_orig",
                "aia_scale_other",
                "Maproj",
                "HmiB2ptr",
            ]
            process_strings = {}
            for p, opts in process.items():
                if p not in allowed_processes:
                    raise ValueError(f"{p} is not one of the allowed processing options: {allowed_processes}")
                process_strings[p] = ",".join([f"{k}={v}" for k, v in opts.items()])
            processes = "|".join([f"{k},{v}" for k, v in process_strings.items()])
            d["process=n"] = f"{d['process=n']}|{processes}"

        if requester is None:
            d["requester"] = notify.split("@")[0]
        elif requester is not False:
            d["requester"] = requester

        query = "?" + urlencode(d)
        return self._server.url_jsoc_fetch + query

    def exp_status(self, requestid):
        """
        Query data export status.

        Parameters
        ----------
        requestid : str
            Request identifier returned by exp_request.

        Returns
        -------
        result : dict
            Dictionary containing the export request status.
        """
        query = f"?{urlencode({'op': 'exp_status', 'requestid': requestid})}"
        req = self._json_request(self._server.url_jsoc_fetch + query)
        return req.data



"""
drms
====

The ``drms`` library provides an easy-to-use interface for accessing HMI, AIA and MDI data with Python.
It uses the publicly accessible JSOC DRMS server by default, but can also be used with local NetDRMS sites.
More information, including a detailed tutorial, is available in the Documentation.

* Homepage: https://github.com/sunpy/drms
* Documentation: https://docs.sunpy.org/projects/drms/en/stable/
"""

import logging
from pathlib import Path

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

from .client import Client, ExportRequest, SeriesInfo
from .config import ServerConfig, register_server
from .exceptions import DrmsError, DrmsExportError, DrmsOperationNotSupported, DrmsQueryError
from .json import HttpJsonClient, HttpJsonRequest, JsocInfoConstants
from .utils import to_datetime
from .version import version as __version__


def _get_bibtex():
    import textwrap

    # Set the bibtex entry to the article referenced in CITATION.rst
    citation_file = Path(__file__).parent / "CITATION.rst"
    # Explicitly specify UTF-8 encoding in case the system's default encoding is problematic
    with citation_file.open(encoding="utf-8") as citation:
        # Extract the first bibtex block:
        ref = citation.read().partition(".. code:: bibtex\n\n")[2]
        lines = ref.split("\n")
        # Only read the lines which are indented
        lines = lines[: [line.startswith("    ") for line in lines].index(False)]
        return textwrap.dedent("\n".join(lines))


__citation__ = __bibtex__ = _get_bibtex()

__all__ = [
    "Client",
    "DrmsError",
    "DrmsExportError",
    "DrmsOperationNotSupported",
    "DrmsQueryError",
    "ExportRequest",
    "HttpJsonClient",
    "HttpJsonRequest",
    "JsocInfoConstants",
    "SeriesInfo",
    "ServerConfig",
    "__bibtex__",
    "__citation__",
    "__version__",
    "logger",
    "register_server",
    "to_datetime",
]



__all__ = [
    "DrmsError",
    "DrmsExportError",
    "DrmsOperationNotSupported",
    "DrmsQueryError",
]


class DrmsError(RuntimeError):
    """
    Unspecified DRMS run-time error.
    """


class DrmsQueryError(DrmsError):
    """
    DRMS query error.
    """


class DrmsExportError(DrmsError):
    """
    DRMS data export error.
    """


class DrmsOperationNotSupported(DrmsError):
    """
    Operation is not supported by DRMS server.
    """



import pytest

from drms.config import ServerConfig, _server_configs, register_server


def test_create_config_basic():
    cfg = ServerConfig(name="TEST")
    valid_keys = ServerConfig._valid_keys
    assert "name" in valid_keys
    assert "encoding" in valid_keys
    for k in valid_keys:
        v = getattr(cfg, k)
        if k == "name":
            assert v == "TEST"
        elif k == "encoding":
            assert v == "latin1"
        else:
            assert v is None
    assert repr(cfg) == "<ServerConfig: TEST>"


def test_create_config_missing_name():
    with pytest.raises(ValueError, match='Server config entry "name" is missing'):
        ServerConfig()


def test_copy_config():
    cfg = ServerConfig(name="TEST")
    assert cfg.name == "TEST"

    cfg2 = cfg.copy()
    assert cfg2 is not cfg
    assert cfg2.name == "TEST"

    cfg.name = "MUH"
    assert cfg.name != cfg2.name


def test_register_server():
    cfg = ServerConfig(name="TEST")

    assert "test" not in _server_configs
    register_server(cfg)
    assert "test" in _server_configs

    del _server_configs["test"]
    assert "test" not in _server_configs


def test_register_server_existing():
    assert "jsoc" in _server_configs
    cfg = ServerConfig(name="jsoc")
    with pytest.raises(RuntimeError):
        register_server(cfg)
    assert "jsoc" in _server_configs


def test_config_jsoc():
    assert "jsoc" in _server_configs

    cfg = _server_configs["jsoc"]
    assert cfg.name.lower() == "jsoc"
    assert isinstance(cfg.encoding, str)
    assert isinstance(cfg.cgi_show_series, str)
    assert isinstance(cfg.cgi_jsoc_info, str)
    assert isinstance(cfg.cgi_jsoc_fetch, str)
    assert isinstance(cfg.cgi_check_address, str)
    assert isinstance(cfg.cgi_show_series_wrapper, str)
    assert isinstance(cfg.show_series_wrapper_dbhost, str)
    assert cfg.http_download_baseurl.startswith("http://")

    baseurl = cfg.cgi_baseurl
    assert baseurl.startswith("http://")
    assert cfg.url_show_series.startswith(baseurl)
    assert cfg.url_jsoc_info.startswith(baseurl)
    assert cfg.url_jsoc_fetch.startswith(baseurl)
    assert cfg.url_check_address.startswith(baseurl)
    assert cfg.url_show_series_wrapper.startswith(baseurl)


def test_config_kis():
    assert "kis" in _server_configs
    cfg = _server_configs["kis"]

    assert cfg.name.lower() == "kis"
    assert isinstance(cfg.encoding, str)

    assert isinstance(cfg.cgi_show_series, str)
    assert isinstance(cfg.cgi_jsoc_info, str)
    assert cfg.cgi_jsoc_fetch is None
    assert cfg.cgi_check_address is None
    assert cfg.cgi_show_series_wrapper is None
    assert cfg.show_series_wrapper_dbhost is None
    assert cfg.http_download_baseurl is None

    baseurl = cfg.cgi_baseurl
    assert baseurl.startswith("http://")
    assert cfg.url_show_series.startswith(baseurl)
    assert cfg.url_jsoc_info.startswith(baseurl)
    assert cfg.url_jsoc_fetch is None
    assert cfg.url_check_address is None
    assert cfg.url_show_series_wrapper is None


@pytest.mark.parametrize(
    ("server_name", "operation", "expected"),
    [
        ("jsoc", "series", True),
        ("jsoc", "info", True),
        ("jsoc", "query", True),
        ("jsoc", "email", True),
        ("jsoc", "export", True),
        ("kis", "series", True),
        ("kis", "info", True),
        ("kis", "query", True),
        ("kis", "email", False),
        ("kis", "export", False),
    ],
)
def test_supported(server_name, operation, expected):
    cfg = _server_configs[server_name]
    assert cfg.check_supported(operation) == expected


@pytest.mark.parametrize(
    ("server_name", "operation"),
    [("jsoc", "bar"), ("kis", "foo")],
)
def test_supported_invalid_operation(server_name, operation):
    cfg = _server_configs[server_name]
    with pytest.raises(ValueError, match="Unknown operation:"):
        cfg.check_supported(operation)


def test_create_config_invalid_key():
    with pytest.raises(ValueError, match="Invalid server config key: foo"):
        ServerConfig(foo="bar")


def test_getset_attr():
    cfg = ServerConfig(name="TEST")
    assert cfg.name == "TEST"
    assert cfg.__dict__ == {"_d": {"encoding": "latin1", "name": "TEST"}}
    with pytest.raises(AttributeError, match="'ServerConfig' object has no attribute 'foo'"):
        _ = cfg.foo
    cfg.name = "NewTest"
    assert cfg.name == "NewTest"
    with pytest.raises(ValueError, match="name config value must be a string"):
        cfg.name = 123
    cfg.__sizeof__ = 127
    assert cfg.__sizeof__ == 127


def test_to_dict():
    cfg = ServerConfig(name="TEST")
    _dict = cfg.to_dict()
    assert isinstance(_dict, dict)
    assert _dict == {"encoding": "latin1", "name": "TEST"}


def test_inbuilt_dir():
    cfg = ServerConfig(name="TEST")
    valid_keys = ServerConfig._valid_keys
    list_attr = dir(cfg)
    assert isinstance(list_attr, list)
    assert set(dir(object)).issubset(set(list_attr))
    assert set(valid_keys).issubset(set(list_attr))
    assert set(cfg.__dict__.keys()).issubset(set(list_attr))



import re
import sys
from urllib.request import Request

import numpy as np
import pandas as pd
from packaging.version import Version

import drms

__all__ = ["create_request_with_header", "to_datetime"]

PD_VERSION = Version(pd.__version__)


def create_request_with_header(url):
    request = Request(url)
    request.add_header("User-Agent", f"drms/{drms.__version__}, python/{sys.version[:5]}")
    return request


def _pd_to_datetime_coerce(arg):
    if PD_VERSION >= Version("2.0.0"):
        return pd.to_datetime(arg, errors="coerce", format="mixed", dayfirst=False)
    return pd.to_datetime(arg, errors="coerce")


def _pd_to_numeric_coerce(arg):
    return pd.to_numeric(arg, errors="coerce")


def _split_arg(arg):
    """
    Split a comma-separated string into a list.
    """
    if isinstance(arg, str):
        return [it for it in re.split(r"[\s,]+", arg) if it]
    return arg


def _extract_series_name(ds):
    """
    Extract series name from record set.
    """
    m = re.match(r"^\s*([\w\.]+).*$", ds)
    return m.group(1) if m is not None else None


def to_datetime(tstr, *, force=False):
    """
    Parse JSOC time strings.

    In general, this is quite complicated, because of the many
    different (non-standard) time strings supported by the DRMS. For
    more (much more!) details on this matter, see
    `Rick Bogart's notes <http://jsoc.stanford.edu/doc/timerep.html>`__.

    The current implementation only tries to convert typical HMI time
    strings, with a format like "%Y.%m.%d_%H:%M:%S_TAI", to an ISO time
    string, that is then parsed by pandas. Note that "_TAI", as well as
    other timezone identifiers like "Z", will not be taken into
    account, so the result will be a naive timestamp without any
    associated timezone.

    If you know the time string format, it might be better calling
    pandas.to_datetime() directly. For handling TAI timestamps, e.g.
    converting between TAI and UTC, the astropy.time package can be
    used.

    Parameters
    ----------
    tstr : str or List[str] or pandas.Series
        Datetime strings.
    force : bool
        Set to True to omit the ``endswith('_TAI')`` check.

    Returns
    -------
    result : pandas.Series or pandas.Timestamp
        Pandas series or a single Timestamp object.
    """
    date = pd.Series(tstr, dtype=object).astype(str)
    if force or date.str.endswith("_TAI").any():
        date = date.str.replace("_TAI", "")
        date = date.str.replace("_", " ")
        if PD_VERSION >= Version("2.0.0"):
            regex = False
        else:
            regex = True
        date = date.str.replace(".", "-", regex=regex, n=2)
    res = _pd_to_datetime_coerce(date)
    res = res.dt.tz_localize(None)
    return res.iloc[0] if (len(res) == 1) and np.isscalar(tstr) else res



import pytest


@pytest.mark.jsoc()
@pytest.mark.remote_data()
def test_series_list_all(jsoc_client):
    slist = jsoc_client.series()
    assert isinstance(slist, list)
    assert "hmi.v_45s" in (s.lower() for s in slist)
    assert "hmi.m_720s" in (s.lower() for s in slist)
    assert "hmi.ic_720s" in (s.lower() for s in slist)
    assert "aia.lev1" in (s.lower() for s in slist)
    assert "aia.lev1_euv_12s" in (s.lower() for s in slist)
    assert "mdi.fd_v" in (s.lower() for s in slist)


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize("schema", ["aia", "hmi", "mdi"])
def test_series_list_schemata(jsoc_client, schema):
    regex = rf"{schema}\."
    slist = jsoc_client.series(regex)
    assert len(slist) > 0
    for sname in slist:
        assert sname.startswith(f"{schema}.")



"""
=========================================
Downloading and plotting a AIA lightcurve
=========================================

This example shows how to download AIA data from JSOC and make a lightcurve plot.
"""

import matplotlib.pyplot as plt

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Some keywords we are interested in; you can use client.keys(series) to get a
# list of all available keywords of a series.

keys = [
    "T_REC",
    "T_OBS",
    "DATAMIN",
    "DATAMAX",
    "DATAMEAN",
    "DATARMS",
    "DATASKEW",
    "DATAKURT",
    "QUALITY",
]

###############################################################################
# Get detailed information about the series. Some keywords from
# aia.lev1_euv_12s are links to keywords in aia.lev1 and unfortunately some
# entries (like note) are missing for linked keywords, so we are using the
# entries from aia.lev1 in this case.

print("Querying series info...")
series_info = client.info("aia.lev1_euv_12s")
series_info_lev1 = client.info("aia.lev1")
for key in keys:
    linkinfo = series_info.keywords.loc[key].linkinfo
    if linkinfo is not None and linkinfo.startswith("lev1->"):
        note_str = series_info_lev1.keywords.loc[key].note
    else:
        note_str = series_info.keywords.loc[key].note
    print(f"{key:>10} : {note_str}")

###############################################################################
# Construct the DRMS query string: "Series[timespan][wavelength]"

qstr = "aia.lev1_euv_12s[2014-01-01T00:00:01Z/365d@1d][335]"

# Get keyword values for the selected timespan and wavelength
print(f"Querying keyword data...\n -> {qstr}")
result = client.query(qstr, key=keys)
print(f" -> {len(result)} lines retrieved.")

# Only use entries with QUALITY==0
result = result[result.QUALITY == 0]
print(f" -> {len(result)} lines after QUALITY selection.")

# Convert T_REC strings to datetime and use it as index for the series
result.index = drms.to_datetime(result.T_REC)

###############################################################################
# Create some simple plots

ax = result[["DATAMIN", "DATAMAX", "DATAMEAN", "DATARMS", "DATASKEW"]].plot(figsize=(8, 10), subplots=True)
ax[0].set_title(qstr, fontsize="medium")
plt.tight_layout()
plt.show()



from collections import OrderedDict

import pandas as pd

import drms


def test_parse_keywords():
    info = [
        {
            "recscope": "variable",
            "units": "none",
            "name": "cparms_sg000",
            "defval": "compress Rice",
            "note": "",
            "type": "string",
        },
        {
            "recscope": "variable",
            "units": "none",
            "name": "mean_bzero",
            "defval": "0",
            "note": "",
            "type": "double",
        },
        {
            "recscope": "variable",
            "units": "none",
            "name": "mean_bscale",
            "defval": "0.25",
            "note": "",
            "type": "double",
        },
        {
            "recscope": "variable",
            "units": "TAI",
            "name": "MidTime",
            "defval": "-4712.01.01_11:59_TAI",
            "note": "Midpoint of averaging interval",
            "type": "time",
        },
    ]
    exp = OrderedDict(
        [
            ("name", ["cparms_sg000", "mean_bzero", "mean_bscale", "MidTime"]),
            ("type", ["string", "double", "double", "time"]),
            ("recscope", ["variable", "variable", "variable", "variable"]),
            ("defval", ["compress Rice", "0", "0.25", "-4712.01.01_11:59_TAI"]),
            ("units", ["none", "none", "none", "TAI"]),
            ("note", ["", "", "", "Midpoint of averaging interval"]),
            ("linkinfo", [None, None, None, None]),
            ("is_time", [False, False, False, True]),
            ("is_integer", [False, False, False, False]),
            ("is_real", [False, True, True, False]),
            ("is_numeric", [False, True, True, False]),
        ],
    )

    exp = pd.DataFrame(data=exp)
    exp.index = exp.pop("name")
    assert drms.SeriesInfo._parse_keywords(info).equals(exp)


def test_parse_links():
    links = [
        {"name": "BHARP", "kind": "DYNAMIC", "note": "Bharp", "target": "hmi.Bharp_720s"},
        {"name": "MHARP", "kind": "DYNAMIC", "note": "Mharp", "target": "hmi.Mharp_720s"},
    ]
    exp = OrderedDict(
        [
            ("name", ["BHARP", "MHARP"]),
            ("target", ["hmi.Bharp_720s", "hmi.Mharp_720s"]),
            ("kind", ["DYNAMIC", "DYNAMIC"]),
            ("note", ["Bharp", "Mharp"]),
        ],
    )
    exp = pd.DataFrame(data=exp)
    exp.index = exp.pop("name")
    assert drms.SeriesInfo._parse_links(links).equals(exp)


def test_parse_segments():
    segments = [
        {
            "type": "int",
            "dims": "VARxVAR",
            "units": "Gauss",
            "protocol": "fits",
            "note": "magnetogram",
            "name": "magnetogram",
        },
        {
            "type": "char",
            "dims": "VARxVAR",
            "units": "Enumerated",
            "protocol": "fits",
            "note": "Mask for the patch",
            "name": "bitmap",
        },
        {
            "type": "int",
            "dims": "VARxVAR",
            "units": "m/s",
            "protocol": "fits",
            "note": "Dopplergram",
            "name": "Dopplergram",
        },
    ]
    exp = OrderedDict(
        [
            ("name", ["magnetogram", "bitmap", "Dopplergram"]),
            ("type", ["int", "char", "int"]),
            ("units", ["Gauss", "Enumerated", "m/s"]),
            ("protocol", ["fits", "fits", "fits"]),
            ("dims", ["VARxVAR", "VARxVAR", "VARxVAR"]),
            ("note", ["magnetogram", "Mask for the patch", "Dopplergram"]),
        ],
    )

    exp = pd.DataFrame(data=exp)
    exp.index = exp.pop("name")
    assert drms.SeriesInfo._parse_segments(segments).equals(exp)


def test_repr():
    info = {
        "primekeys": ["CarrRot", "CMLon"],
        "retention": 1800,
        "tapegroup": 1,
        "archive": 1,
        "primekeysinfo": [],
        "unitsize": 1,
        "note": "Temporal averages of HMI Vgrams over 1/3 CR",
        "dbindex": [None],
        "links": [],
        "segments": [],
        "keywords": [],
    }
    assert repr(drms.SeriesInfo(info)) == "<SeriesInfo>"
    assert repr(drms.SeriesInfo(info, name="hmi")) == "<SeriesInfo: hmi>"



import sys
import argparse


def main():
    import drms

    args = parse_args(sys.argv[1:])
    client = drms.Client(server=args.server, email=args.email)
    drms.logger.info(f"client: {client}")


def parse_args(args):
    import drms

    parser = argparse.ArgumentParser(description="drms, access HMI, AIA and MDI data with python")
    parser.add_argument(
        "--version",
        action="version",
        version=f"drms v{drms.__version__}",
        help="show package version and exit",
    )
    parser.add_argument("--email", help="email address for data export requests")
    parser.add_argument("server", nargs="?", default="jsoc", help="DRMS server, default is JSOC")

    return parser.parse_args(args)



import pandas as pd
import pytest

from drms.utils import _extract_series_name, _pd_to_datetime_coerce, _pd_to_numeric_coerce, _split_arg


@pytest.mark.parametrize(
    ("in_obj", "expected"),
    [
        ("", []),
        ("asd", ["asd"]),
        ("aa,bb,cc", ["aa", "bb", "cc"]),
        ("aa, bb, cc", ["aa", "bb", "cc"]),
        (" aa,bb,  cc, dd", ["aa", "bb", "cc", "dd"]),
        ("aa,\tbb,cc, dd ", ["aa", "bb", "cc", "dd"]),
        ([], []),
        (["a", "b", "c"], ["a", "b", "c"]),
        (("a", "b", "c"), ["a", "b", "c"]),
    ],
)
def test_split_arg(in_obj, expected):
    res = _split_arg(in_obj)
    assert len(res) == len(expected)
    for i in range(len(res)):
        assert res[i] == expected[i]


@pytest.mark.parametrize(
    ("ds_string", "expected"),
    [
        ("hmi.v_45s", "hmi.v_45s"),
        ("hmi.v_45s[2010.05.01_TAI]", "hmi.v_45s"),
        ("hmi.v_45s[2010.05.01_TAI/365d@1d]", "hmi.v_45s"),
        ("hmi.v_45s[2010.05.01_TAI/365d@1d][?QUALITY>=0?]", "hmi.v_45s"),
        ("hmi.v_45s[2010.05.01_TAI/1d@6h]{Dopplergram}", "hmi.v_45s"),
    ],
)
def test_extract_series(ds_string, expected):
    assert _extract_series_name(ds_string) == expected


@pytest.mark.parametrize(
    ("arg", "exp"),
    [
        (pd.Series(["1.0", "2", -3]), pd.Series([1.0, 2.0, -3.0])),
        (pd.Series(["1.0", "apple", -3]), pd.Series([1.0, float("nan"), -3.0])),
    ],
)
def test_pd_to_numeric_coerce(arg, exp):
    assert _pd_to_numeric_coerce(arg).equals(exp)


@pytest.mark.parametrize(
    ("arg", "exp"),
    [
        (
            pd.Series(["2016-04-01 00:00:00", "2016-04-01 06:00:00"]),
            pd.Series([pd.Timestamp("2016-04-01 00:00:00"), pd.Timestamp("2016-04-01 06:00:00")]),
        ),
    ],
)
def test_pd_to_datetime_coerce(arg, exp):
    assert _pd_to_datetime_coerce(arg).equals(exp)



"""
==================================
List all keywords for a HMI series
==================================

This example shows how to display the keywords for a HMI series.
"""

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Query series info

series_info = client.info("hmi.v_45s")

###############################################################################
# Print keyword info

print(f"Listing keywords for {series_info.name}:\n")
for keyword in sorted(series_info.keywords.index):
    keyword_info = series_info.keywords.loc[keyword]
    print(keyword)
    print(f"  type ....... {keyword_info.type} ")
    print(f"  recscope ... {keyword_info.recscope} ")
    print(f"  defval ..... {keyword_info.defval} ")
    print(f"  units ...... {keyword_info.units} ")
    print(f"  note ....... {keyword_info.note} ")



"""
=============================================
Downloading and plotting solar modes with HMI
=============================================

This example shows how to download HMI data from JSOC and make a plot of the solar modes.
"""

import matplotlib.pyplot as plt
import numpy as np

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Construct the DRMS query string: "Series[timespan][wavelength]"

qstr = "hmi.v_sht_modes[2024.09.19_00:00:00_TAI]"

# TODO: Add text here.
segname = "m6"  # 'm6', 'm18' or 'm36'

# Send request to the DRMS server
print(f"Querying keyword data...\n -> {qstr}")
result, filenames = client.query(qstr, key=["T_START", "T_STOP", "LMIN", "LMAX", "NDT"], seg=segname)
print(f" -> {len(result)} lines retrieved.")

# Use only the first line of the query result
result = result.iloc[0]
fname = f"http://jsoc.stanford.edu{filenames[segname][0]}"

# Read the data segment
print(f"Reading data from {fname}...")
a = np.genfromtxt(fname)

# For column names, see appendix of Larson & Schou (2015SoPh..290.3221L)
l = a[:, 0].astype(int)
n = a[:, 1].astype(int)
nu = a[:, 2] / 1e3
if a.shape[1] in [24, 48, 84]:
    # tan(gamma) present
    sig_offs = 5
elif a.shape[1] in [26, 50, 86]:
    # tan(gamma) not present
    sig_offs = 6
snu = a[:, sig_offs + 2] / 1e3

###############################################################################
# Plot the zoomed in on lower l
fig, ax = plt.subplots(1, 1, figsize=(11, 7))
ax.set_title(
    f"Time = {result.T_START} ... {result.T_STOP}, L = {result.LMIN} ... {result.LMAX}, NDT = {result.NDT}",
    fontsize="medium",
)
for ni in np.unique(n):
    idx = n == ni
    ax.plot(l[idx], nu[idx], "b.-")
ax.set_xlim(0, 120)
ax.set_ylim(0.8, 4.5)
ax.set_xlabel("Harmonic degree")
ax.set_ylabel("Frequency [mHz]")
fig.tight_layout()

###############################################################################
# Plot the zoomed in on higher l, n <= 20, with errors

fig, ax = plt.subplots(1, 1, figsize=(11, 7))
ax.set_title(
    f"Time = {result.T_START} ... {result.T_STOP}, L = {result.LMIN} ... {result.LMAX}, NDT = {result.NDT}",
    fontsize="medium",
)
for ni in np.unique(n):
    if ni <= 20:
        idx = n == ni
        ax.plot(l[idx], nu[idx], "b.", ms=3)
        if ni < 10:
            ax.plot(l[idx], nu[idx] + 1000 * snu[idx], "g")
            ax.plot(l[idx], nu[idx] - 1000 * snu[idx], "g")
        else:
            ax.plot(l[idx], nu[idx] + 500 * snu[idx], "r")
            ax.plot(l[idx], nu[idx] - 500 * snu[idx], "r")
ax.legend(
    loc="upper right",
    handles=[
        plt.Line2D([0], [0], color="r", label="500 sigma"),
        plt.Line2D([0], [0], color="g", label="1000 sigma"),
    ],
)
ax.set_xlim(-5, 305)
ax.set_ylim(0.8, 4.5)
ax.set_xlabel("Harmonic degree")
ax.set_ylabel("Frequency [mHz]")
fig.tight_layout()

plt.show()



import sys

import pytest

from drms.main import main, parse_args


def helper(args, name, expected):
    args = parse_args(args)
    assert getattr(args, name) == expected


def test_version():
    with pytest.raises(SystemExit):
        helper(["--version"], "version", expected=True)


def test_server():
    helper(["fake_url_server"], "server", "fake_url_server")
    helper([], "server", "jsoc")


def test_email():
    helper(["--email", "fake@gmail.com"], "email", "fake@gmail.com")
    helper([], "email", None)


def test_main_empty():
    sys.argv = [
        "drms",
    ]
    main()



"""
=========================================
Downloading and plotting a HMI lightcurve
=========================================

This example shows how to download HMI data from JSOC and make a lightcurve plot.
"""

import matplotlib.pyplot as plt

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Construct the DRMS query string: "Series[timespan]"

qstr = "hmi.ic_720s[2010.05.01_TAI-2016.04.01_TAI@6h]"

# Send request to the DRMS server
print(f"Querying keyword data...\n -> {qstr}")
result = client.query(qstr, key=["T_REC", "DATAMEAN", "DATARMS"])
print(f" -> {int(len(result))} lines retrieved.")

###############################################################################
# Now to plot the image.

# Convert T_REC strings to datetime and use it as index for the series
result.index = drms.to_datetime(result.pop("T_REC"))

# Note: DATARMS contains the standard deviation, not the RMS!
t = result.index
avg = result.DATAMEAN / 1e3
std = result.DATARMS / 1e3

# Create plot
fig, ax = plt.subplots(1, 1, figsize=(15, 7))
ax.set_title(qstr, fontsize="medium")
ax.fill_between(
    t,
    avg + std,
    avg - std,
    edgecolor="none",
    facecolor="b",
    alpha=0.3,
    interpolate=True,
)
ax.plot(t, avg, color="b")
ax.set_xlabel("Time")
ax.set_ylabel("Disk-averaged continuum intensity [kDN/s]")
fig.tight_layout()

plt.show()



"""
============================================
Downloading and plotting a HMI synoptic data
============================================

This example shows how to download HMI synoptic data from JSOC and make a plot.
"""

import matplotlib.pyplot as plt

from astropy.io import fits

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# Construct the DRMS query string: "Series[Carrington rotation]"

qstr = "hmi.synoptic_mr_720s[2150]"

# Send request to the DRMS server
print(f"Querying keyword data...\n -> {qstr}")
segname = "synopMr"
results, filenames = client.query(qstr, key=drms.JsocInfoConstants.all, seg=segname)
print(f" -> {len(results)} lines retrieved.")

# Use only the first line of the query result
results = results.iloc[0]
fname = f"http://jsoc.stanford.edu{filenames[segname][0]}"

# Read the data segment
# Note: HTTP downloads get cached in ~/.astropy/cache/downloads
print(f"Reading data from {fname}...")
a = fits.getdata(fname)
ny, nx = a.shape

###############################################################################
# Now to plot the image.

# Convert pixel to world coordinates using WCS keywords
xmin = (1 - results.CRPIX1) * results.CDELT1 + results.CRVAL1
xmax = (nx - results.CRPIX1) * results.CDELT1 + results.CRVAL1
ymin = (1 - results.CRPIX2) * results.CDELT2 + results.CRVAL2
ymax = (ny - results.CRPIX2) * results.CDELT2 + results.CRVAL2

# Convert to Carrington longitude
xmin = results.LON_LAST - xmin
xmax = results.LON_LAST - xmax

# Compute the plot extent used with imshow
extent = (
    xmin - abs(results.CDELT1) / 2,
    xmax + abs(results.CDELT1) / 2,
    ymin - abs(results.CDELT2) / 2,
    ymax + abs(results.CDELT2) / 2,
)

# Aspect ratio for imshow in respect to the extent computed above
aspect = abs((xmax - xmin) / nx * ny / (ymax - ymin))

# Create plot
fig, ax = plt.subplots(1, 1, figsize=(13.5, 6))
ax.set_title(f"{qstr}, Time: {results.T_START} ... {results.T_STOP}", fontsize="medium")
ax.imshow(
    a,
    vmin=-300,
    vmax=300,
    origin="lower",
    interpolation="nearest",
    cmap="gray",
    extent=extent,
    aspect=aspect,
)
ax.invert_xaxis()
ax.set_xlabel("Carrington longitude")
ax.set_ylabel("Sine latitude")
fig.tight_layout()

plt.show()



"""
====================================
Downloading data as a tar collection
====================================

This example shows how to submit a data export request using the 'url-tar'
method, which provides a single TAR archive containing all requested files.

Here we use this method to download data from the
'hmi.rdvflows_fd15_frame' series, which stores directories of text files
for each record. This is currently the only way to download directory
data segments using the Python DRMS client. The export protocol in this
case is 'as-is'. You might change the protocol to 'fits', if you are
downloading FITS files instead of text files.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: "Series[Carrington rotation][Carrington longitude]{data segments}"

qstr = "hmi.rdvflows_fd15_frame[2150][360]{Ux, Uy}"
print(f"Data export query:\n  {qstr}\n")

# Submit export request using the 'url-tar' method, protocol default: 'as-is'
print("Submitting export request...")
result = client.export(qstr, method="url-tar", email=email)

# Print request URL.
print(f"\nRequest URL: {result.request_url}")
print(f"{len(result.urls)} file(s) available for download.\n")

# Download selected files.
dr = result.download(out_dir)
print("Download finished.")
print(f'\nDownloaded file:\n  "{dr.download[0]}"\n')



import numpy as np
import pandas as pd
import pytest

import drms

data_tai = [
    ("2010.05.01_TAI", pd.Timestamp("2010-05-01 00:00:00")),
    ("2010.05.01_00:00_TAI", pd.Timestamp("2010-05-01 00:00:00")),
    ("2010.05.01_00:00:00_TAI", pd.Timestamp("2010-05-01 00:00:00")),
    ("2010.05.01_01:23:45_TAI", pd.Timestamp("2010-05-01 01:23:45")),
    ("2013.12.21_23:32_TAI", pd.Timestamp("2013-12-21 23:32:00")),
    ("2013.12.21_23:32:34_TAI", pd.Timestamp("2013-12-21 23:32:34")),
]
data_tai_in = [data[0] for data in data_tai]
data_tai_out = pd.Series([data[1] for data in data_tai])


@pytest.mark.parametrize(("time_string", "expected"), data_tai)
def test_tai_string(time_string, expected):
    assert drms.to_datetime(time_string) == expected


@pytest.mark.parametrize(
    ("time_string", "expected"),
    [
        ("2010-05-01T00:00Z", pd.Timestamp("2010-05-01 00:00:00")),
        ("2010-05-01T00:00:00Z", pd.Timestamp("2010-05-01 00:00:00")),
        ("2010-05-01T01:23:45Z", pd.Timestamp("2010-05-01 01:23:45")),
        ("2013-12-21T23:32Z", pd.Timestamp("2013-12-21 23:32:00")),
        ("2013-12-21T23:32:34Z", pd.Timestamp("2013-12-21 23:32:34")),
        ("2010-05-01 00:00Z", pd.Timestamp("2010-05-01 00:00:00")),
        ("2010-05-01 00:00:00Z", pd.Timestamp("2010-05-01 00:00:00")),
        ("2010-05-01 01:23:45Z", pd.Timestamp("2010-05-01 01:23:45")),
        ("2013-12-21 23:32Z", pd.Timestamp("2013-12-21 23:32:00")),
        ("2013-12-21 23:32:34Z", pd.Timestamp("2013-12-21 23:32:34")),
    ],
)
def test_z_string(time_string, expected):
    assert drms.to_datetime(time_string) == expected


@pytest.mark.xfail(reason="pandas does not support leap seconds")
@pytest.mark.parametrize(
    ("time_string", "expected"),
    [
        ("2012-06-30T23:59:60Z", "2012-06-30 23:59:60"),
        ("2015-06-30T23:59:60Z", "2015-06-30 23:59:60"),
        ("2016-12-31T23:59:60Z", "2016-12-31 23:59:60"),
    ],
)
def test_z_leap_string(time_string, expected):
    assert drms.to_datetime(time_string) == expected


@pytest.mark.parametrize(
    ("time_string", "expected"),
    [
        ("2013.12.21_23:32:34_TAI", pd.Timestamp("2013-12-21 23:32:34")),
        ("2013.12.21_23:32:34_UTC", pd.Timestamp("2013-12-21 23:32:34")),
        ("2013.12.21_23:32:34Z", pd.Timestamp("2013-12-21 23:32:34")),
    ],
)
def test_force_string(time_string, expected):
    assert drms.to_datetime(time_string, force=True) == expected


@pytest.mark.parametrize(
    ("time_series", "expected"),
    [
        (data_tai_in, data_tai_out),
        (pd.Series(data_tai_in), data_tai_out),
        (tuple(data_tai_in), data_tai_out),
        (np.array(data_tai_in), data_tai_out),
    ],
)
def test_time_series(time_series, expected):
    assert drms.to_datetime(time_series).equals(expected)


data_invalid = [
    ("2010.05.01_TAI", False),
    ("2010.05.01_00:00_TAI", False),
    ("", True),
    ("1600", True),
    ("foo", True),
    ("2013.12.21_23:32:34_TAI", False),
]
data_invalid_in = [data[0] for data in data_invalid]
data_invalid_out = pd.Series([data[1] for data in data_invalid])


@pytest.mark.parametrize(("time_string", "expected"), data_invalid)
def test_corner_case(time_string, expected):
    assert pd.isna(drms.to_datetime(time_string)) == expected
    assert isinstance(drms.to_datetime([]), pd.Series)
    assert drms.to_datetime([]).empty


@pytest.mark.parametrize(
    ("time_series", "expected"),
    [
        (data_invalid_in, data_invalid_out),
        (pd.Series(data_invalid_in), data_invalid_out),
        (tuple(data_invalid_in), data_invalid_out),
        (np.array(data_invalid_in), data_invalid_out),
    ],
)
def test_corner_case_series(time_series, expected):
    assert pd.isna(drms.to_datetime(time_series)).equals(expected)



import re

import pytest

import drms


@pytest.mark.parametrize(
    "symbol",
    [
        "__bibtex__",
        "__citation__",
        "__version__",
        "client",
        "Client",
        "config",
        "DrmsError",
        "DrmsExportError",
        "DrmsOperationNotSupported",
        "DrmsQueryError",
        "exceptions",
        "ExportRequest",
        "HttpJsonClient",
        "HttpJsonRequest",
        "JsocInfoConstants",
        "json",
        "main",
        "Path",
        "register_server",
        "SeriesInfo",
        "ServerConfig",
        "to_datetime",
        "utils",
        "version",
    ],
)
def test_symbols(symbol):
    assert symbol in dir(drms)


def test_version():
    assert isinstance(drms.__version__, str)
    version = drms.__version__.split("+")[0]
    # Check to make sure it isn't empty
    assert version
    # To match the 0.6 in 0.6.dev3 or v0.6
    m = re.match(r"v*\d+\.\d+\.*", version)
    assert m is not None


def test_bibtex():
    assert isinstance(drms.__citation__, str)
    m = re.match(r".*Glogowski2019.*", drms.__citation__)
    assert m is not None



"""
=================
Exporting a movie
=================

This example shows how to export movies from image data, using the 'mp4'
protocol.

The 'mp4' protocol accepts additional protocol arguments, like color
table, color scaling or pixel binning. For a list of available color
tables, see http://jsoc.stanford.edu/ajax/exportdata.html and select the
MP4 protocol.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: "Series[timespan]{segment}"

qstr = "hmi.M_720s[2025.01.29_23:12:00_TAI/5d@1h]{magnetogram}"
print(f"Data export query:\n  {qstr}\n")

###############################################################################
# Arguments for 'mp4' protocol

mp4_args = {
    "ct": "grey.sao",  # color table
    "min": -1500,  # min value
    "max": 1500,  # max value
    "scaling": "mag",  # color scaling
    "size": 8,  # binning (1 -> 4k, 2 -> 2k, 4 -> 1k, 8 -> 512)
}

# Submit export request using the 'mp4' protocol with custom protocol_args
print("Submitting export request...")
result = client.export(qstr, protocol="mp4", protocol_args=mp4_args, email=email)
result.wait(sleep=10)

# Print request URL.
print(f"\nRequest URL: {result.request_url}")
print(f"{len(result.urls)} file(s) available for download.\n")

# Download movie file only: index=0
result.download(out_dir, index=0)
print("Download finished.")
print(f"\nDownload directory:\n  {out_dir.resolve()}\n")



"""
=========================
Exporting data as 'as-is'
=========================

This example shows how to submit an 'url_quick' / 'as-is' data export request
to JSOC and how to download the requested files.

The export protocol 'as-is', should be preferred over other protocols,
because it minimizes the server load. The only reason to use
protocol='fits' is, when keywords in the FITS header are really needed.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: "Series[harpnum][timespan]{data segments}"

qstr = "hmi.sharp_720s[7451][2020.09.27_00:00:00_TAI]{continuum, magnetogram, field}"
print(f"Data export query:\n  {qstr}\n")

# Submit export request, defaults to method='url_quick' and protocol='as-is'
print("Submitting export request...")
result = client.export(qstr, email=email)
print(f"{int(len(result.urls))} file(s) available for download.\n")

# Download selected files.
result.download(out_dir)
print("Download finished.")
print(f'\nDownload directory:\n  "{out_dir.resolve()}"\n')



"""
==============================
Getting the urls of a download
==============================

This example prints the download URLs for files returned from an 'as-is' data
export request.

Note that there is no "Request URL" for method 'url_quick'.
"""

import os

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

# This example requires a registered export email address. You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.
email = os.environ["JSOC_EMAIL"]

###############################################################################
# Construct the DRMS query string: "Series[timespan][wavelength]"

qstr = "hmi.ic_720s[2015.01.01_00:00:00_TAI/10d@1d]{continuum}"

# Submit export request, defaults to method='url_quick' and protocol='as-is'
print(f"Data export query:\n  {qstr}\n")
print("Submitting export request...")
result = client.export(qstr, email=email)
print(f"{len(result.urls)} file(s) available for download.\n")

# Print download URLs.
for _, row in result.urls[["record", "url"]].iterrows():
    print(f"REC(ORD): {row.record}")
    print(f"URL: {row.url}\n")



import pytest

import drms


@pytest.mark.parametrize(
    "exception_class",
    [drms.DrmsError, drms.DrmsQueryError, drms.DrmsExportError, drms.DrmsOperationNotSupported],
)
def test_exception_class(exception_class):
    with pytest.raises(RuntimeError):
        raise exception_class
    with pytest.raises(drms.DrmsError):
        raise exception_class



"""
===================================
Exporting data with cutout requests
===================================

This example shows how to submit a data export request with a cutout request using ``im_patch``.
"""

import os
from pathlib import Path

import drms

###############################################################################
# First we will create a `drms.Client`, using the JSOC baseurl.

client = drms.Client()

###############################################################################
# This example requires a registered export email address You can register
# JSOC exports at: http://jsoc.stanford.edu/ajax/register_email.html
# You must supply your own email.

email = os.environ["JSOC_EMAIL"]

# Create download directory if it does not exist yet.
out_dir = Path("downloads")
if not out_dir.exists():
    Path(out_dir).mkdir(parents=True)

###############################################################################
# Construct the DRMS query string: ``"Series[timespan][wavelength]{data segments}"``

qstr = "aia.lev1_euv_12s[2025-01-01T04:33:30.000/1m@12s][171]{image}"
print(f"Data export query:\n  {qstr}\n")

###############################################################################
# Construct the dictionary specifying that we want to request a cutout.
# This is done via the ``im_patch`` command.
# We request a 345.6 arcsecond cutout (on both sides) centered on the coordinate (-517.2, -246) arcseconds as defined in the helioprojective frame of SDO at time ``t_ref``.
# The ``t`` controls whether tracking is disabled (``1``) or enabled (``0``).
# ``r`` controls the use of sub-pixel registration.
# ``c`` controls whether off-limb pixels are filled with NaNs.
# For additional details about ``im_patch``, see the `documentation <http://jsoc.stanford.edu/doxygen_html/group__im__patch.html>`_.
process = {
    "im_patch": {
        "t_ref": "2025-01-01T04:33:30.000",
        "t": 0,
        "r": 0,
        "c": 0,
        "locunits": "arcsec",
        "boxunits": "arcsec",
        "x": -517.2,
        "y": -246,
        "width": 345.6,
        "height": 345.6,
    },
}

# Submit export request using the 'fits' protocol
print("Submitting export request...")
result = client.export(
    qstr,
    method="url",
    protocol="fits",
    email=email,
    process=process,
)

# Print request URL.
print(f"\nRequest URL: {result.request_url}")
print(f"{int(len(result.urls))} file(s) available for download.\n")

# Download selected files.
result.wait()
result.download(out_dir)
print("Download finished.")
print(f'\nDownload directory:\n  "{out_dir.resolve()}"\n')



import pytest


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize(
    ("series", "pkeys", "segments"),
    [
        ("hmi.v_45s", ["T_REC", "CAMERA"], ["Dopplergram"]),
        ("hmi.m_720s", ["T_REC", "CAMERA"], ["magnetogram"]),
        ("hmi.v_sht_2drls", ["LMIN", "NACOEFF"], ["split", "rot", "err"]),
    ],
)
def test_series_info_basic(jsoc_client, series, pkeys, segments):
    si = jsoc_client.info(series)
    assert si.name.lower() == series
    for k in pkeys:
        assert k in si.primekeys
        assert k in si.keywords.index
    for s in segments:
        assert s in si.segments.index


@pytest.mark.jsoc()
@pytest.mark.remote_data()
@pytest.mark.parametrize(
    ("series", "pkeys"),
    [
        ("hmi.v_45s", ["T_REC", "CAMERA"]),
        ("hmi.m_720s", ["T_REC", "CAMERA"]),
        ("hmi.v_sht_2drls", ["LMIN", "NACOEFF"]),
        ("aia.lev1", ["T_REC", "FSN"]),
        ("aia.lev1_euv_12s", ["T_REC", "WAVELNTH"]),
        ("aia.response", ["T_START", "WAVE_STR"]),
        ("iris.lev1", ["T_OBS", "FSN"]),
        ("mdi.fd_m_lev182", ["T_REC"]),
    ],
)
def test_series_primekeys(jsoc_client, series, pkeys):
    pkey_list = jsoc_client.pkeys(series)
    key_list = jsoc_client.keys(series)
    for k in pkeys:
        assert k in pkey_list
        assert k in key_list
